#!/usr/bin/env python

import os
import sys
from distutils import dir_util
from scipy_distutils.misc_util   import get_path, default_config_dict, dot_join
from scipy_distutils.misc_util   import dict_append, get_build_temp
from scipy_distutils.misc_util   import SourceFilter
from scipy_distutils.core        import Extension
from scipy_distutils.system_info import get_info

from distutils.sysconfig         import get_python_lib

from config_pygist import config_pygist

def get_special_dirs(plat):
    if plat in ['aix4', 'aix5','sunos5']:
        return ['/usr/local/lib']
    elif plat in ['linux2','cygwin']:
        return ['/usr/lib']
    return []

gistsource = ["src/gist/gist.c",
              "src/gist/tick.c",
              "src/gist/tick60.c",
              "src/gist/engine.c",
              "src/gist/gtext.c",
              "src/gist/draw.c",
              "src/gist/draw0.c",
              "src/gist/clip.c",
              "src/gist/gread.c",
              "src/gist/gcntr.c",
              "src/gist/hlevel.c",
              "src/gist/ps.c",
              "src/gist/cgm.c",
              "src/gist/eps.c",
              "src/gist/style.c",
              "src/gist/xfancy.c",
              "src/gist/xbasic.c"]

cygwin_unixsource = ["src/play/unix/dir.c",
                     "src/play/unix/files.c",
                     "src/play/unix/pathnm.c",
                     "src/play/unix/slinks.c",
                     "src/play/unix/stdinit.c",
                     "src/play/unix/uevent.c",
                     "src/play/unix/uinbg.c",
                     "src/play/unix/usernm.c"]

unixsource = ["src/play/unix/dir.c",
              "src/play/unix/files.c",
              "src/play/unix/fpuset.c",
              "src/play/unix/pathnm.c",
              "src/play/unix/timew.c",
              "src/play/unix/uevent.c",
              "src/play/unix/ugetc.c",
              "src/play/unix/umain.c",
              "src/play/unix/usernm.c",
              "src/play/unix/slinks.c"]

x11source = ["src/play/x11/colors.c",
             "src/play/x11/connect.c",
             "src/play/x11/cursors.c",
             "src/play/x11/errors.c",
             "src/play/x11/events.c",
             "src/play/x11/fills.c",
             "src/play/x11/fonts.c",
             "src/play/x11/images.c",
             "src/play/x11/lines.c",
             "src/play/x11/pals.c",
             "src/play/x11/pwin.c",
             "src/play/x11/resource.c",
             "src/play/x11/rgbread.c",
             "src/play/x11/textout.c",
             "src/play/x11/rect.c",
             "src/play/x11/clips.c",
             "src/play/x11/points.c"]

winsource = ["src/play/win/pscr.c",
             "src/play/win/pals.c",
             "src/play/win/ptext.c",
             "src/play/win/pfill.c",
             "src/play/win/pcell.c",
             "src/play/win/pmin.c",
             "src/play/win/plines.c",
             "src/play/win/prect.c",
             "src/play/win/points.c",
             "src/play/win/cursors.c",
             "src/play/win/pwin.c",
             "src/play/win/timew.c",
             "src/play/win/clips.c",
             "src/play/win/getdc.c",
             "src/play/win/files.c",
             "src/play/win/usernm.c",
             "src/play/win/pathnm.c"]

cygwin_winsource = ["src/play/win/pscr.c",
                    "src/play/win/pals.c",
                    "src/play/win/ptext.c",
                    "src/play/win/pfill.c",
                    "src/play/win/pcell.c",
                    "src/play/win/pmin.c",
                    "src/play/win/plines.c",
                    "src/play/win/prect.c",
                    "src/play/win/points.c",
                    "src/play/win/cursors.c",
                    "src/play/win/pwin.c",
                    "src/play/win/timew.c",
                    "src/play/win/clips.c",
                    "src/play/win/getdc.c"]

allsource = ["src/play/all/hash.c",
             "src/play/all/hash0.c",
             "src/play/all/mm.c",
             "src/play/all/alarms.c",
             "src/play/all/pstrcpy.c",
             "src/play/all/pstrncat.c",
             "src/play/all/p595.c",
             "src/play/all/bitrev.c",
             "src/play/all/bitlrot.c",
             "src/play/all/bitmrot.c"]

def filter_playsource(sources,local_path):
    if sys.platform == 'win32':
        playsource = winsource + allsource
    elif sys.platform == 'cygwin':
        playsource = cygwin_unixsource + cygwin_winsource + allsource
    else:
        playsource = unixsource + x11source + allsource
    return [os.path.join(local_path,n) for n in playsource]

def getallparams(gistpath,local_path,config_path):
    x11_info = get_info('x11')
    extra_compile_args = ['-DGISTPATH="\\"' + gistpath + '\\""' ]

    extra_link_args = []
    if sys.platform in ['win32','cygwin']:
        extra_compile_args.append("-DWINDOWS")
        extra_compile_args.append("-mwindows")
        extra_link_args.append("-mwindows")
        libraries = []
    else:
        libraries = x11_info.get('libraries',['X11'])
    if sys.platform=='cygwin':
        extra_compile_args.append("-DCYGWIN")

    include_dirs = ['src/gist', 'src/play', 'src/play/unix' ]

    library_dirs = [os.path.join(local_path,x) for x in ['.','src']]
    library_dirs.extend(x11_info.get('library_dirs',[]))
    library_dirs.extend(get_special_dirs(sys.platform))

    include_dirs = [os.path.join(local_path,x) for x in include_dirs]
    include_dirs.extend(x11_info.get('include_dirs',[]))

    if 1:
        inputfile = open(os.path.join(config_path,"Make.cfg"))
        lines = inputfile.readlines()
        inputfile.close()
        for line in lines:
            if line[:8]=="MATHLIB=":
                mathlib = line[8:-1] #removing the \n
                # remove the -l
                mathlib = mathlib[2:]
                libraries.append(mathlib)
            if line[:9]=="NO_EXP10=":
                no_exp10 = line[9:-1] # removing \n
                if no_exp10: extra_compile_args.append(no_exp10)
            if line[:5]=="XINC=":
                xinc = line[5:-1] # removing \n
                if xinc and sys.platform not in ['cygwin','win32']:
                    # remove the -I
                    xinc = xinc[2:]
                    if xinc: include_dirs.append(xinc)
            if line[:5]=="XLIB=":
                xlib = line[5:-1] # removing \n
                if xlib and sys.platform not in ['cygwin','win32']:
                    # remove the -L
                    xlib = xlib[2:]
                    library_dirs.append(xlib)    

    return include_dirs, library_dirs, libraries, \
               extra_compile_args, extra_link_args
    

def configuration(parent_package='',parent_path=None):
    """
       This will install *.gs and *.gp files to
       'site-packages/scipy/xplt/gistdata' 
    """
    from glob import glob
    local_path = get_path(__name__,parent_path)
    config_path = os.path.join(get_build_temp(),'config_pygist')
    dir_util.mkpath(config_path,verbose=1)

    conf = config_pygist(local_path,config_path)
    # Look to see if compiler is set on command line and add it 
    #    This is repeating code, but I'm not sure how to avoid it
    #    As this gets run before overall setup does.
    #    This is needed so that compiler can be over-ridden from the
    #    platform default in the configuration section of xplt.
    for arg in sys.argv[1:]:
        if arg[:11] == '--compiler=':
            conf.compiler = arg[11:]
            break
        if arg[:2] == '-c':
            conf.compiler = arg[2:]
            break
    # Generate Make.cfg and config.h:
    conf.run()

    package = 'xplt'        
    xplt_path = os.path.join(parent_package,'xplt')
    config = default_config_dict(package,parent_package)

    all_playsource = glob(os.path.join(local_path,'src','play','*','*.c')) + \
      glob(os.path.join(local_path,'src','play','*.h'))
    playsource = SourceFilter(filter_playsource, all_playsource, local_path)

    gistpath = os.path.join(get_python_lib(1),xplt_path,"gistdata")
    gistpath = gistpath.replace("\\",r"\\\\")

    gistC = os.path.join(local_path,'pygist','gistCmodule.c')
    sources = [os.path.join(local_path,x) for x in gistsource]
    sources = [gistC] + sources + [playsource]

    include_dirs, library_dirs, libraries, \
                  extra_compile_args, extra_link_args \
                  = getallparams(gistpath,local_path,config_path)
    include_dirs.insert(0,os.path.dirname(conf.config_h))

    ext_arg = {'name':dot_join(parent_package,package,'gistC'),
               'sources':sources,
               'include_dirs':include_dirs,
               'library_dirs':library_dirs,
               'libraries':libraries,
               'extra_compile_args':extra_compile_args,
               'extra_link_args':extra_link_args,
               'depends':[os.path.join(local_path,'src')]}
    ext = Extension (**ext_arg)
    config['ext_modules'].append(ext)

    file_ext = ['*.gs','*.gp', '*.ps', '*.help']
    xplt_files = [glob(os.path.join(local_path,'gistdata',x)) for x in file_ext]
    xplt_files += [glob(os.path.join(local_path,'src','g',x)) for x in file_ext]
    xplt_files = reduce(lambda x,y:x+y,xplt_files,[])
    data_path = os.path.join(xplt_path,'gistdata')
    config['data_files'].extend( [(data_path,xplt_files)])

    return config

if __name__ == '__main__':    
    from scipy_distutils.core import setup
    setup(**configuration(parent_path=''))



# Author: Travis Oliphant
# 1999 -- 2002

import sigtools
import scipy.special as special
import scipy.linalg as linalg
from scipy.fftpack import fft, ifft, ifftshift, fft2, ifft2
from scipy_base import polyadd, polymul, polydiv, polysub, \
     roots, poly, polyval, polyder, cast, asarray, isscalar
import types
import scipy
from scipy.stats import mean
import Numeric
from Numeric import array, arange, where, sqrt, rank, zeros, NewAxis, argmax
from scipy_base.fastumath import *

_modedict = {'valid':0, 'same':1, 'full':2}
_boundarydict = {'fill':0, 'pad':0, 'wrap':2, 'circular':2, 'symm':1, 'symmetric':1, 'reflect':4}
                                                                            
def _valfrommode(mode):
    try:
        val = _modedict[mode]
    except KeyError:
        if mode not in [0,1,2]:
            raise ValueError, "Acceptable mode flags are 'valid' (0), 'same' (1), or 'full' (2)."
        val = mode
    return val

def _bvalfromboundary(boundary):
    try:
        val = _boundarydict[boundary] << 2
    except KeyError:
        if val not in [0,1,2] :
            raise ValueError, "Acceptable boundary flags are 'fill', 'wrap' (or 'circular'), \n  and 'symm' (or 'symmetric')."
        val = boundary << 2
    return val


def correlate(in1, in2, mode='full'):
    """Cross-correlate two N-dimensional arrays.

  Description:

     Cross-correlate in1 and in2 with the output size determined by mode.

  Inputs:

    in1 -- an N-dimensional array.
    in2 -- an array with the same number of dimensions as in1.
    mode -- a flag indicating the size of the output
            'valid'  (0): The output consists only of those elements that
                            do not rely on the zero-padding.
            'same'   (1): The output is the same size as the largest input
                            centered with respect to the 'full' output.
            'full'   (2): The output is the full discrete linear
                            cross-correlation of the inputs. (Default)

  Outputs:  (out,)

    out -- an N-dimensional array containing a subset of the discrete linear
           cross-correlation of in1 with in2.
 
    """
    # Code is faster if kernel is smallest array.
    volume = asarray(in1)
    kernel = asarray(in2)
    if rank(volume) == rank(kernel) == 0:
        return volume*kernel
    if (Numeric.product(kernel.shape) > Numeric.product(volume.shape)):
        temp = kernel
        kernel = volume
        volume = temp
        del temp

    val = _valfrommode(mode)

    return sigtools._correlateND(volume, kernel, val)

def convolve(in1, in2, mode='full'):
    """Convolve two N-dimensional arrays.

  Description:

     Convolve in1 and in2 with output size determined by mode.

  Inputs:

    in1 -- an N-dimensional array.
    in2 -- an array with the same number of dimensions as in1.
    mode -- a flag indicating the size of the output
            'valid'  (0): The output consists only of those elements that
                            are computed by scaling the larger array with all
                            the values of the smaller array.
            'same'   (1): The output is the same size as the largest input
                            centered with respect to the 'full' output.
            'full'   (2): The output is the full discrete linear convolution
                            of the inputs. (Default)

  Outputs:  (out,)

    out -- an N-dimensional array containing a subset of the discrete linear
           convolution of in1 with in2.

    """
    volume = asarray(in1)
    kernel = asarray(in2)
    if rank(volume) == rank(kernel) == 0:
        return volume*kernel
    if (Numeric.product(kernel.shape) > Numeric.product(volume.shape)):
        temp = kernel
        kernel = volume
        volume = temp
        del temp

    slice_obj = [slice(None,None,-1)]*len(kernel.shape)
    val = _valfrommode(mode)
    
    return sigtools._correlateND(volume,kernel[slice_obj],val)

def order_filter(a, domain, order):
    """Perform an order filter on an N-dimensional array.
    
  Description:

    Perform an order filter on the array in.  The domain argument acts as a
    mask centered over each pixel.  The non-zero elements of domain are
    used to select elements surrounding each input pixel which are placed
    in a list.   The list is sorted, and the output for that pixel is the
    element corresponding to rank in the sorted list.
    
  Inputs:

    in -- an N-dimensional input array.
    domain -- a mask array with the same number of dimensions as in.  Each
              dimension should have an odd number of elements.
    rank -- an non-negative integer which selects the element from the
            sorted list (0 corresponds to the largest element, 1 is the
            next largest element, etc.)

  Output: (out,)

    out -- the results of the order filter in an array with the same
           shape as in.
          
    """
    domain = asarray(domain)
    size = domain.shape
    for k in range(len(size)):
        if (size[k] % 2) != 1:
            raise ValueError, "Each dimension of domain argument should have an odd number of elements."
    return sigtools._orderfilterND(a, domain, rank)
   

def medfilt(volume,kernel_size=None):
    """Perform a median filter on an N-dimensional array.

  Description:

    Apply a median filter to the input array using a local window-size
    given by kernel_size.

  Inputs:

    in -- An N-dimensional input array.
    kernel_size -- A scalar or an N-length list giving the size of the
                   median filter window in each dimension.  Elements of
                   kernel_size should be odd.  If kernel_size is a scalar,
                   then this scalar is used as the size in each dimension.

  Outputs: (out,)

    out -- An array the same size as input containing the median filtered
           result.
  
    """
    volume = asarray(volume)
    if kernel_size is None:
        kernel_size = [3] * len(volume.shape)
    kernel_size = asarray(kernel_size)
    if len(kernel_size.shape) == 0:
        kernel_size = [kernel_size.toscalar()] * len(volume.shape)
    kernel_size = asarray(kernel_size)

    for k in range(len(volume.shape)):
        if (kernel_size[k] % 2) != 1:
            raise ValueError, "Each element of kernel_size should be odd." 

    domain = Numeric.ones(kernel_size)

    numels = Numeric.product(kernel_size)
    order = int(numels/2)
    return sigtools._order_filterND(volume,domain,order)


def wiener(im,mysize=None,noise=None):
    """Perform a wiener filter on an N-dimensional array.

  Description:

    Apply a wiener filter to the N-dimensional array in.

  Inputs:

    in -- an N-dimensional array.
    kernel_size -- A scalar or an N-length list giving the size of the
                   median filter window in each dimension.  Elements of
                   kernel_size should be odd.  If kernel_size is a scalar,
                   then this scalar is used as the size in each dimension.
    noise -- The noise-power to use.  If None, then noise is estimated as
             the average of the local variance of the input.

  Outputs: (out,)

    out -- Wiener filtered result with the same shape as in.

    """
    im = asarray(im)
    if mysize is None:
        mysize = [3] * len(im.shape)
    mysize = asarray(mysize);

    # Estimate the local mean
    lMean = correlate(im,Numeric.ones(mysize),1) / Numeric.product(mysize)

    # Estimate the local variance
    lVar = correlate(im**2,Numeric.ones(mysize),1) / Numeric.product(mysize) - lMean**2

    # Estimate the noise power if needed.
    if noise==None:
        noise = mean(Numeric.ravel(lVar))

    res = (im - lMean)
    res *= (1-noise / lVar)
    res += lMean
    out = where(lVar < noise, lMean, res)

    return out
    

def convolve2d(in1, in2, mode='full', boundary='fill', fillvalue=0):
    """Conolve two 2-dimensional arrays.

  Description:

     Convolve in1 and in2 with output size determined by mode and boundary
     conditions determined by boundary and fillvalue.

  Inputs:

    in1 -- a 2-dimensional array.
    in2 -- a 2-dimensional array.
    mode -- a flag indicating the size of the output
            'valid'  (0): The output consists only of those elements that
                            do not rely on the zero-padding.
            'same'   (1): The output is the same size as the input centered
                            with respect to the 'full' output.
            'full'   (2): The output is the full discrete linear convolution
                            of the inputs. (*Default*)
    boundary -- a flag indicating how to handle boundaries
                'fill' : pad input arrays with fillvalue. (*Default*)
                'wrap' : circular boundary conditions.
                'symm' : symmetrical boundary conditions.
    fillvalue -- value to fill pad input arrays with (*Default* = 0)

  Outputs:  (out,)

    out -- a 2-dimensional array containing a subset of the discrete linear
           convolution of in1 with in2.

    """
    val = _valfrommode(mode)
    bval = _bvalfromboundary(boundary)
        
    return sigtools._convolve2d(in1,in2,1,val,bval,fillvalue)

def correlate2d(in1, in2, mode='full', boundary='fill', fillvalue=0):
    """Cross-correlate two 2-dimensional arrays.

  Description:

     Cross correlate in1 and in2 with output size determined by mode
     and boundary conditions determined by boundary and fillvalue.

  Inputs:

    in1 -- a 2-dimensional array.
    in2 -- a 2-dimensional array.
    mode -- a flag indicating the size of the output
            'valid'  (0): The output consists only of those elements that
                            do not rely on the zero-padding.
            'same'   (1): The output is the same size as the input centered
                            with respect to the 'full' output.
            'full'   (2): The output is the full discrete linear convolution
                            of the inputs. (*Default*)
    boundary -- a flag indicating how to handle boundaries
                'fill' : pad input arrays with fillvalue. (*Default*)
                'wrap' : circular boundary conditions.
                'symm' : symmetrical boundary conditions.
    fillvalue -- value to fill pad input arrays with (*Default* = 0)

  Outputs:  (out,)

    out -- a 2-dimensional array containing a subset of the discrete linear
           cross-correlation of in1 with in2.

    """
    val = _valfrommode(mode)
    bval = _bvalfromboundary(boundary)
        
    return sigtools._convolve2d(in1, in2, 0,val,bval,fillvalue)

def medfilt2d(input, kernel_size=3):
    """Median filter two 2-dimensional arrays.

  Description:

    Apply a median filter to the input array using a local window-size
    given by kernel_size (must be odd).

  Inputs:

    in -- An 2 dimensional input array.
    kernel_size -- A scalar or an length-2 list giving the size of the
                   median filter window in each dimension.  Elements of
                   kernel_size should be odd.  If kernel_size is a scalar,
                   then this scalar is used as the size in each dimension.

  Outputs: (out,)

    out -- An array the same size as input containing the median filtered
           result.
    """
    image = asarray(input)
    if kernel_size is None:
        kernel_size = [3] * 2
    kernel_size = asarray(kernel_size)
    if len(kernel_size.shape) == 0:
        kernel_size = [kernel_size.toscalar()] * 2
    kernel_size = asarray(kernel_size)

    for size in kernel_size:
        if (size % 2) != 1:
            raise ValueError, "Each element of kernel_size should be odd." 

    return sigtools._medfilt2d(image, kernel_size)

def remez(numtaps, bands, desired, weight=None, Hz=1, type='bandpass',
          maxiter=25, grid_density=16):
    """Calculate the minimax optimal filter using Remez exchange algorithm.
    
  Description:

    Calculate the filter-coefficients for the finite impulse response
    (FIR) filter whose transfer function minimizes the maximum error
    between the desired gain and the realized gain in the specified bands
    using the remez exchange algorithm.

  Inputs:

    numtaps -- The desired number of taps in the filter.
    bands -- A montonic sequence containing the band edges.  All elements
             must be non-negative and less than 1/2 the sampling frequency
             as given by Hz.
    desired -- A sequency half the size of bands containing the desired gain
               in each of the specified bands
    weight -- A relative weighting to give to each band region.
    type --- The type of filter:
             'bandpass' : flat response in bands.
             'differentiator' : frequency proportional response in bands.

  Outputs: (out,)

    out -- A rank-1 array containing the coefficients of the optimal
           (in a minimax sense) filter.
    
    """
    # Convert type
    try:
        tnum = {'bandpass':1, 'differentiator':2}[type]
    except KeyError:
        raise ValueError, "Type must be 'bandpass', or 'differentiator'"

    # Convert weight
    if weight is None:
        weight = [1] * len(desired)

    bands = bands.copy()
    return sigtools._remez(numtaps, bands, desired, weight, tnum, Hz,
                           maxiter, grid_density)

def lfilter(b, a, x, axis=-1, zi=None):
    """Filter data along one-dimension with an IIR or FIR filter.

  Description

    Filter a data sequence, x, using a digital filter.  This works for many
    fundamental data types (including Object type).  The filter is a direct
    form II transposed implementation of the standard difference equation
    (see "Algorithm"). 

  Inputs:

    b -- The numerator coefficient vector in a 1-D sequence.
    a -- The denominator coefficient vector in a 1-D sequence.  If a[0]
         is not 1, then both a and b are normalized by a[0].
    x -- An N-dimensional input array.
    axis -- The axis of the input data array along which to apply the
            linear filter. The filter is applied to each subarray along
            this axis (*Default* = -1)
    zi -- Initial conditions for the filter delays.  It is a vector
          (or array of vectors for an N-dimensional input) of length
          max(len(a),len(b)).  If zi=None or is not given then initial
          rest is assumed.  SEE signal.lfiltic for more information.

  Outputs: (y, {zf})

    y -- The output of the digital filter.
    zf -- If zi is None, this is not returned, otherwise, zf holds the
          final filter delay values.

  Algorithm:

    The filter function is implemented as a direct II transposed structure.
    This means that the filter implements

    y[n] = b[0]*x[n] + b[1]*x[n-1] + ... + b[nb]*x[n-nb]
                     - a[1]*y[n-1] + ... + a[na]*y[n-na]

    using the following difference equations:

    y[m] = b[0]*x[m] + z[0,m-1]
    z[0,m] = b[1]*x[m] + z[1,m-1] - a[1]*y[m]
    ...
    z[n-3,m] = b[n-2]*x[m] + z[n-2,m-1] - a[n-2]*y[m]
    z[n-2,m] = b[n-1]*x[m] - a[n-1]*y[m]

    where m is the output sample number and n=max(len(a),len(b)) is the
    model order.

    The rational transfer function describing this filter in the
    z-transform domain is
                                -1               -nb
                    b[0] + b[1]z  + ... + b[nb] z
            Y(z) = ---------------------------------- X(z)
                                -1               -na
                    a[0] + a[1]z  + ... + a[na] z
                    
    """
    if isscalar(a):
	a = [a]
    if zi is None:
        return sigtools._linear_filter(b, a, x, axis)
    else:
        return sigtools._linear_filter(b, a, x, axis, zi)

def lfiltic(b,a,y,x=None):
    """Given a linear filter (b,a) and initial conditions on the output y
    and the input x, return the inital conditions on the state vector zi
    which is used by lfilter to generate the output given the input.

    If M=len(b)-1 and N=len(a)-1.  Then, the initial conditions are given
    in the vectors x and y as 

    x = {x[-1],x[-2],...,x[-M]}
    y = {y[-1],y[-2],...,y[-N]}

    If x is not given, its inital conditions are assumed zero.
    If either vector is too short, then zeros are added
      to achieve the proper length.

    The output vector zi contains

    zi = {z_0[-1], z_1[-1], ..., z_K-1[-1]}  where K=max(M,N).
    """
    N = Numeric.size(a)-1
    M = Numeric.size(b)-1
    K = max(M,N)
    y = asarray(y)
    zi = zeros(K,y.typecode())
    if x is None:
        x = zeros(M,y.typecode())
    else:
        x = asarray(x)
        L = Numeric.size(x)
        if L < M:
            x = r_[x,zeros(M-L)]
    L = Numeric.size(y)
    if L < N:
        y = r_[y,zeros(N-L)]

    for m in range(M):
        zi[m] = Numeric.sum(b[m+1:]*x[:M-m])

    for m in range(N):
        zi[m] -= Numeric.sum(a[m+1:]*y[:N-m])

    return zi
    

def boxcar(M,sym=1):
    """The M-point boxcar window.
    """
    return Numeric.ones(M,Numeric.Float)

def triang(M,sym=1):
    """The M-point triangular window.
    """
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2
    if not sym and not odd:
        M = M + 1        
    n = arange(1,int((M+1)/2)+1)
    if M % 2 == 0:
        w = (2*n-1.0)/M
        w = scipy.r_[w, w[::-1]]
    else:
        w = 2*n/(M+1.0)
        w = scipy.r_[w, w[-2::-1]]

    if not sym and not odd:
        w = w[:-1]
    return w

def parzen(M,sym=1):
    """The M-point Parzen window
    """
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2
    if not sym and not odd:
        M = M+1    
    n = Numeric.arange(-(M-1)/2.0,(M-1)/2.0+0.5,1.0)
    na = extract(n < -(M-1)/4.0, n)
    nb = extract(abs(n) <= (M-1)/4.0, n)
    wa = 2*(1-abs(na)/(M/2.0))**3.0
    wb = 1-6*(abs(nb)/(M/2.0))**2.0 + 6*(abs(nb)/(M/2.0))**3.0
    w = scipy.r_[wa,wb,wa[::-1]]
    if not sym and not odd:
        w = w[:-1]
    return w

def bohman(M,sym=1):
    """The M-point Bohman window
    """
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2
    if not sym and not odd:
        M = M+1    
    fac = abs(linspace(-1,1,M)[1:-1])
    w = (1 - fac)* cos(pi*fac) + 1.0/pi*sin(pi*fac)
    w = scipy.r_[0,w,0]    
    if not sym and not odd:
        w = w[:-1]
    return w

def blackman(M,sym=1):
    """The M-point Blackman window.
    """
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2
    if not sym and not odd:
        M = M+1
    n = arange(0,M)
    w = 0.42-0.5*cos(2.0*pi*n/(M-1)) + 0.08*cos(4.0*pi*n/(M-1))
    if not sym and not odd:
        w = w[:-1]
    return w

def nuttall(M,sym=1):
    """A minimum 4-term Blackman-Harris window according to Nuttall.
    """
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2
    if not sym and not odd:
        M = M+1    
    a = [0.3635819, 0.4891775, 0.1365995, 0.0106411]
    n = arange(0,M)
    fac = n*2*pi/(M-1.0)
    w = a[0] - a[1]*cos(fac) + a[2]*cos(2*fac) - a[3]*cos(3*fac)
    if not sym and not odd:
        w = w[:-1]    
    return w

def blackmanharris(M,sym=1):
    """The M-point minimum 4-term Blackman-Harris window.
    """
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2
    if not sym and not odd:
        M = M+1        
    a = [0.35875, 0.48829, 0.14128, 0.01168];
    n = arange(0,M)
    fac = n*2*pi/(M-1.0)
    w = a[0] - a[1]*cos(fac) + a[2]*cos(2*fac) - a[3]*cos(3*fac)
    if not sym and not odd:
        w = w[:-1]    
    return w

    
def bartlett(M,sym=1):
    """The M-point Bartlett window.
    """
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2
    if not sym and not odd:
        M = M+1    
    n = arange(0,M)
    w = where(Numeric.less_equal(n,(M-1)/2.0),2.0*n/(M-1),2.0-2.0*n/(M-1))
    if not sym and not odd:
        w = w[:-1]
    return w

def hanning(M,sym=1):
    """The M-point Hanning window.
    """
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2
    if not sym and not odd:
        M = M+1        
    n = arange(0,M)
    w = 0.5-0.5*cos(2.0*pi*n/(M-1))
    if not sym and not odd:
        w = w[:-1]
    return w

def barthann(M,sym=1):
    """Return the M-point modified Bartlett-Hann window.
    """
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2
    if not sym and not odd:
        M = M+1            
    n = arange(0,M)
    fac = abs(n/(M-1.0)-0.5)
    w = 0.62 - 0.48*fac + 0.38*cos(2*pi*fac)
    if not sym and not odd:
        w = w[:-1]
    return w    

def hamming(M,sym=1):
    """The M-point Hamming window.
    """
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2
    if not sym and not odd:
        M = M+1        
    n = arange(0,M)    
    w = 0.54-0.46*cos(2.0*pi*n/(M-1))
    if not sym and not odd:
        w = w[:-1]
    return w

def kaiser(M,beta,sym=1):
    """Returns a Kaiser window of length M with shape parameter beta.
    """
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2
    if not sym and not odd:
        M = M+1    
    n = arange(0,M)
    alpha = (M-1)/2.0
    w = special.i0(beta * sqrt(1-((n-alpha)/alpha)**2.0))/special.i0(beta)
    if not sym and not odd:
        w = w[:-1]
    return w

def gaussian(M,std,sym=1):
    """Returns a Gaussian window of length M with standard-deviation std.
    """
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2        
    if not sym and not odd:
        M = M + 1
    n = arange(0,M)-(M-1.0)/2.0
    sig2 = 2*std*std
    w = exp(-n**2 / sig2)
    if not sym and not odd:
        w = w[:-1]
    return w

def general_gaussian(M,p,sig,sym=1):
    """Returns a window with a generalized Gaussian shape.

    exp(-0.5*(x/sig)**(2*p))

    half power point is at (2*log(2)))**(1/(2*p))*sig
    """
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2
    if not sym and not odd:
        M = M+1        
    n = arange(0,M)-(M-1.0)/2.0
    w = exp(-0.5*(n/sig)**(2*p))
    if not sym and not odd:
        w = w[:-1]
    return w


def slepian(M,width,sym=1):
    if (M*width > 27.38):
        raise ValueError, "Cannot reliably obtain slepian sequences for"\
              " M*width > 27.38."
    if M < 1:
        return Numeric.array([])
    if M == 1:
        return Numeric.ones(1,'d')
    odd = M % 2
    if not sym and not odd:
        M = M+1

    twoF = width/2.0
    alpha = (M-1)/2.0
    m = arange(0,M)-alpha
    n = m[:,NewAxis]
    k = m[NewAxis,:]
    AF = twoF*special.sinc(twoF*(n-k))
    [lam,vec] = linalg.eig(AF)
    ind = argmax(abs(lam))
    w = abs(vec[:,ind])
    w = w / max(w)
    
    if not sym and not odd:
        w = w[:-1]
    return w
            

def hilbert(x, N=None):
    """Return the hilbert transform of x of length N.
    """
    x = asarray(x)
    if N is None:
        N = len(x)
    if N <=0:
        raise ValueError, "N must be positive."
    if scipy.iscomplexobj(x):
        print "Warning: imaginary part of x ignored."
        x = scipy.real(x)
    Xf = fft(x,N,axis=0)
    h = Numeric.zeros(N)
    if N % 2 == 0:
        h[0] = h[N/2] = 1
        h[1:N/2] = 2
    else:
        h[0] = 1
        h[1:(N+1)/2] = 2

    if len(x.shape) > 1:
        h = h[:,Numeric.NewAxis]
    x = ifft(Xf*h)
    return x

def hilbert2(x,N=None):
    """Return the '2-D' hilbert transform of x of length N.
    """
    x = asarray(x)
    x = asarray(x)
    if N is None:
        N = x.shape
    if len(N) < 2:
        if N <=0:
            raise ValueError, "N must be positive."
        N = (N,N)
    if scipy.iscomplexobj(x):
        print "Warning: imaginary part of x ignored."
        x = scipy.real(x)
    print N
    Xf = fft2(x,N,axes=(0,1))
    h1 = Numeric.zeros(N[0],'d')
    h2 = Numeric.zeros(N[1],'d')
    for p in range(2):
        h = eval("h%d"%(p+1))
        N1 = N[p]
        if N1 % 2 == 0:
            h[0] = h[N1/2] = 1
            h[1:N1/2] = 2
        else:
            h[0] = 1
            h[1:(N1+1)/2] = 2
        exec("h%d = h" % (p+1), globals(), locals())

    h = h1[:,NewAxis] * h2[NewAxis,:]
    k = len(x.shape)
    while k > 2:
        h = h[:,Numeric.NewAxis]
        k -= 1
    x = ifft2(Xf*h,axes=(0,1))
    return x
    

def cmplx_sort(p):
    "sort roots based on magnitude."
    p = asarray(p)
    if scipy.iscomplexobj(p):
        indx = Numeric.argsort(abs(p))
    else:
        indx = Numeric.argsort(p)
    return Numeric.take(p,indx), indx

def unique_roots(p,tol=1e-3,rtype='min'):
    """Determine the unique roots and their multiplicities in two lists

    Inputs:

      p -- The list of roots
      tol --- The tolerance for two roots to be considered equal.
      rtype --- How to determine the returned root from the close 
                  ones:  'max': pick the maximum
                         'min': pick the minimum
                         'avg': average roots
    Outputs: (pout, mult)

      pout -- The list of sorted roots
      mult -- The multiplicity of each root
    """
    if rtype in ['max','maximum']:
        comproot = scipy.max
    elif rtype in ['min','minimum']:
        comproot = scipy.min
    elif rtype in ['avg','mean']:
        comproot = scipy.mean
    p = asarray(p)*1.0
    tol = abs(tol)
    p, indx = cmplx_sort(p)
    pout = []
    mult = []
    indx = -1
    curp = p[0] + 5*tol
    sameroots = []
    for k in range(len(p)):
        tr = p[k]
        if abs(tr-curp) < tol:
            sameroots.append(tr)
            curp = comproot(sameroots)
            pout[indx] = curp
            mult[indx] += 1
        else:
            pout.append(tr)
            curp = tr
            sameroots = [tr]            
            indx += 1
            mult.append(1)
    return array(pout), array(mult)

from scipy_base import real_if_close, atleast_1d


def invres(r,p,k,tol=1e-3,rtype='avg'):
    """Compute b(s) and a(s) from partial fraction expansion: r,p,k

    If M = len(b) and N = len(a)
    
            b(s)     b[0] x**(M-1) + b[1] x**(M-2) + ... + b[M-1] 
    H(s) = ------ = ----------------------------------------------
            a(s)     a[0] x**(N-1) + a[1] x**(N-2) + ... + a[N-1]

             r[0]       r[1]             r[-1]
         = -------- + -------- + ... + --------- + k(s)
           (s-p[0])   (s-p[1])         (s-p[-1])

    If there are any repeated roots (closer than tol), then the partial
    fraction expansion has terms like

            r[i]      r[i+1]              r[i+n-1]
          -------- + ----------- + ... + ----------- 
          (s-p[i])  (s-p[i])**2          (s-p[i])**n

    See also:  residue, poly, polyval, unique_roots
    """
    extra = k
    p, indx = cmplx_sort(p)
    r = Numeric.take(r,indx)
    pout, mult = unique_roots(p,tol=tol,rtype=rtype)
    p = []
    for k in range(len(pout)):
        p.extend([pout[k]]*mult[k])
    a = atleast_1d(poly(p))
    if len(extra) > 0:
        b = polymul(extra,a)
    else:
        b = [0]
    indx = 0
    for k in range(len(pout)):
        temp = []
        for l in range(len(pout)):
            if l != k:
                temp.extend([pout[l]]*mult[l])
        for m in range(mult[k]):
            t2 = temp[:]
            t2.extend([pout[k]]*(mult[k]-m-1))
            b = polyadd(b,r[indx]*poly(t2))
            indx += 1
    b = real_if_close(b)
    while Numeric.allclose(b[0], 0, rtol=1e-14) and (b.shape[-1] > 1):
        b = b[1:]
    return b, a

from scipy import factorial
def residue(b,a,tol=1e-3,rtype='avg'):
    """Compute partial-fraction expansion of b(s) / a(s).

    If M = len(b) and N = len(a)
    
            b(s)     b[0] s**(M-1) + b[1] s**(M-2) + ... + b[M-1] 
    H(s) = ------ = ----------------------------------------------
            a(s)     a[0] s**(N-1) + a[1] s**(N-2) + ... + a[N-1]

             r[0]       r[1]             r[-1]
         = -------- + -------- + ... + --------- + k(s)
           (s-p[0])   (s-p[1])         (s-p[-1])

    If there are any repeated roots (closer than tol), then the partial
    fraction expansion has terms like 

            r[i]      r[i+1]              r[i+n-1]
          -------- + ----------- + ... + ----------- 
          (s-p[i])  (s-p[i])**2          (s-p[i])**n

    See also:  invres, poly, polyval, unique_roots
    """

    b,a = map(asarray,(b,a))
    k,b = polydiv(b,a)
    p = roots(a)
    r = p*0.0
    pout, mult = unique_roots(p,tol=tol,rtype=rtype)
    p = []
    for n in range(len(pout)):
        p.extend([pout[n]]*mult[n])
    p = asarray(p)
    # Compute the residue from the general formula
    indx = 0
    for n in range(len(pout)):
        bn = b.copy()
        pn = []
        for l in range(len(pout)):
            if l != n:
                pn.extend([pout[l]]*mult[l])
        an = atleast_1d(poly(pn))
        # bn(s) / an(s) is (s-po[n])**Nn * b(s) / a(s) where Nn is
        # multiplicity of pole at po[n]
        sig = mult[n]
        for m in range(sig,0,-1):
            if sig > m:
                # compute next derivative of bn(s) / an(s)
                term1 = polymul(polyder(bn,1),an)
                term2 = polymul(bn,polyder(an,1))
                bn = polysub(term1,term2)
                an = polymul(an,an)                
            r[indx+m-1] = polyval(bn,pout[n]) / polyval(an,pout[n]) \
                          / factorial(sig-m)
        indx += sig
    return r, p, k

def residuez(b,a,tol=1e-3,rtype='avg'):
    """Compute partial-fraction expansion of b(z) / a(z).

    If M = len(b) and N = len(a)
    
            b(z)     b[0] + b[1] z**(-1) + ... + b[M-1] z**(-M+1)
    H(z) = ------ = ----------------------------------------------
            a(z)     a[0] + a[1] z**(-1) + ... + a[N-1] z**(-N+1)

                 r[0]                   r[-1]
         = --------------- + ... + ---------------- + k[0] + k[1]z**(-1) ...
           (1-p[0]z**(-1))         (1-p[-1]z**(-1))

    If there are any repeated roots (closer than tol), then the partial
    fraction expansion has terms like 

               r[i]              r[i+1]                    r[i+n-1]
          -------------- + ------------------ + ... + ------------------ 
          (1-p[i]z**(-1))  (1-p[i]z**(-1))**2         (1-p[i]z**(-1))**n

    See also:  invresz, poly, polyval, unique_roots
    """
    b,a = map(asarray,(b,a))
    gain = a[0]
    brev, arev = b[::-1],a[::-1]
    krev,brev = polydiv(brev,arev)
    if krev == []:
        k = []
    else:
        k = krev[::-1]    
    b = brev[::-1]
    p = roots(a)
    r = p*0.0
    pout, mult = unique_roots(p,tol=tol,rtype=rtype)
    p = []
    for n in range(len(pout)):
        p.extend([pout[n]]*mult[n])
    p = asarray(p)
    # Compute the residue from the general formula (for discrete-time)
    #  the polynomial is in z**(-1) and the multiplication is by terms
    #  like this (1-p[i] z**(-1))**mult[i].  After differentiation,
    #  we must divide by (-p[i])**(m-k) as well as (m-k)!
    indx = 0
    for n in range(len(pout)):
        bn = brev.copy()
        pn = []
        for l in range(len(pout)):
            if l != n:
                pn.extend([pout[l]]*mult[l])
        an = atleast_1d(poly(pn))[::-1]
        # bn(z) / an(z) is (1-po[n] z**(-1))**Nn * b(z) / a(z) where Nn is
        # multiplicity of pole at po[n] and b(z) and a(z) are polynomials.
        sig = mult[n]
        for m in range(sig,0,-1):
            if sig > m:
                # compute next derivative of bn(s) / an(s)
                term1 = polymul(polyder(bn,1),an)
                term2 = polymul(bn,polyder(an,1))
                bn = polysub(term1,term2)
                an = polymul(an,an)                
            r[indx+m-1] = polyval(bn,1.0/pout[n]) / polyval(an,1.0/pout[n]) \
                          / factorial(sig-m) / (-pout[n])**(sig-m)
        indx += sig
    return r/gain, p, k

def invresz(r,p,k,tol=1e-3,rtype='avg'):
    """Compute b(z) and a(z) from partial fraction expansion: r,p,k

    If M = len(b) and N = len(a)
    
            b(z)     b[0] + b[1] z**(-1) + ... + b[M-1] z**(-M+1)
    H(z) = ------ = ----------------------------------------------
            a(z)     a[0] + a[1] z**(-1) + ... + a[N-1] z**(-N+1)

                 r[0]                   r[-1]
         = --------------- + ... + ---------------- + k[0] + k[1]z**(-1) ...
           (1-p[0]z**(-1))         (1-p[-1]z**(-1))

    If there are any repeated roots (closer than tol), then the partial
    fraction expansion has terms like 

               r[i]              r[i+1]                    r[i+n-1]
          -------------- + ------------------ + ... + ------------------ 
          (1-p[i]z**(-1))  (1-p[i]z**(-1))**2         (1-p[i]z**(-1))**n

    See also:  residuez, poly, polyval, unique_roots
    """
    extra = asarray(k)
    p, indx = cmplx_sort(p)
    r = Numeric.take(r,indx)
    pout, mult = unique_roots(p,tol=tol,rtype=rtype)
    p = []
    for k in range(len(pout)):
        p.extend([pout[k]]*mult[k])
    a = atleast_1d(poly(p))
    if len(extra) > 0:
        b = polymul(extra,a)
    else:
        b = [0]
    indx = 0
    brev = asarray(b)[::-1]
    for k in range(len(pout)):  
        temp = []
        # Construct polynomial which does not include any of this root
        for l in range(len(pout)):
            if l != k:
                temp.extend([pout[l]]*mult[l])
        for m in range(mult[k]):
            t2 = temp[:]
            t2.extend([pout[k]]*(mult[k]-m-1))
            brev = polyadd(brev,(r[indx]*poly(t2))[::-1])
            indx += 1
    b = real_if_close(brev[::-1])
    return b, a


def get_window(window,Nx,fftbins=1):
    """Return a window of length Nx and type window.

    If fftbins is 1, create a "periodic" window ready to use with ifftshift
    and be multiplied by the result of an fft (SEE ALSO fftfreq). 

    Window types:  boxcar, triang, blackman, hamming, hanning, bartlett,
                   parzen, bohman, blackmanharris, nuttall, barthann,
                   kaiser (needs beta), gaussian (needs std),
                   general_gaussian (needs power, width),
                   slepian (needs width)

    If the window requires no parameters, then it can be a string.
    If the window requires parameters, the window argument should be a tuple
        with the first argument the string name of the window, and the next
        arguments the needed parameters.
    If window is a floating point number, it is interpreted as the beta
        parameter of the kaiser window.
    """

    sym = not fftbins
    try:
        beta = float(window)
    except (TypeError, ValueError):
        args = ()
        if isinstance(window, types.TupleType):
            winstr = window[0]
            if len(window) > 1:
                args = window[1:]
        elif isinstance(window, types.StringType):
            if window in ['kaiser', 'ksr', 'gaussian', 'gauss', 'gss',
                        'general gaussian', 'general_gaussian',
                        'general gauss', 'general_gauss', 'ggs']:
                raise ValueError, "That window needs a parameter -- pass a tuple"
            else:
                winstr = window
                
        if winstr in ['blackman', 'black', 'blk']:
            winfunc = blackman
        elif winstr in ['triangle', 'triang', 'tri']:
            winfunc = triang
        elif winstr in ['hamming', 'hamm', 'ham']:
            winfunc = hamming
        elif winstr in ['bartlett', 'bart', 'brt']:
            winfunc = bartlett
        elif winstr in ['hanning', 'hann', 'han']:
            winfunc = hanning
        elif winstr in ['blackmanharris', 'blackharr','bkh']:
            winfun = blackmanharris
        elif winstr in ['parzen', 'parz', 'par']:
            winfun = parzen
        elif winstr in ['bohman', 'bman', 'bmn']:
            winfunc = bohman
        elif winstr in ['nuttall', 'nutl', 'nut']:
            winfunc = nuttall
        elif winstr in ['barthann', 'brthan', 'bth']:
            winfunc = barthann
            
        elif winstr in ['kaiser', 'ksr']:
            winfunc = kaiser
        elif winstr in ['gaussian', 'gauss', 'gss']:
            winfunc = gaussian
        elif winstr in ['general gaussian', 'general_gaussian',
                        'general gauss', 'general_gauss', 'ggs']:
            winfunc = general_gaussian
        elif winstr in ['boxcar', 'box', 'ones']:
            winfunc = boxcar
        elif winstr in ['slepian', 'slep', 'optimal', 'dss']:
            winfunc = slepian
        else:
            raise ValueError, "Unknown window type."

        params = (Nx,)+args + (sym,)
    else:
        winfunc = kaiser
        params = (Nx,beta,sym)

    return winfunc(*params)
        

def resample(x,num,t=None,axis=0,window=None):
    """Resample to num samples using Fourier method along the given axis.

    The resampled signal starts at the same value of x but is sampled
    with a spacing of len(x) / num * (spacing of x).  Because a Fourier method
    is used, the signal is assumed periodic.

    Window controls a Fourier-domain window that tapers the Fourier spectrum
    before zero-padding to aleviate ringing in the resampled values for
    sampled signals you didn't intend to be interpreted as band-limited.

    If window is a string then use the named window.  If window is a float, then
    it represents a value of beta for a kaiser window.  If window is a tuple,
    then the first component is a string representing the window, and the next
    arguments are parameters for that window.

    Possible windows are:
           'blackman'       ('black',   'blk')
           'hamming'        ('hamm',    'ham')
           'bartlett'       ('bart',    'brt')
           'hanning'        ('hann',    'han')
           'kaiser'         ('ksr')             # requires parameter (beta)
           'gaussian'       ('gauss',   'gss')  # requires parameter (std.)
           'general gauss'  ('general', 'ggs')  # requires two parameters
                                                      (power, width)

    The first sample of the returned vector is the same as the first sample of the
        input vector, the spacing between samples is changed from dx to
        dx * len(x) / num

    If t is not None, then it represents the old sample positions, and the new
       sample positions will be returned as well as the new samples.
    """
    x = asarray(x)
    X = fft(x,axis=axis)
    Nx = x.shape[axis]
    if window is not None:
        W = ifftshift(get_window(window,Nx))
        newshape = ones(len(x.shape))
        newshape[axis] = len(W)
        W.shape = newshape
        X = X*W
    sl = [slice(None)]*len(x.shape)
    newshape = list(x.shape)
    newshape[axis] = num
    N = int(Numeric.minimum(num,Nx))
    Y = Numeric.zeros(newshape,'D')
    sl[axis] = slice(0,(N+1)/2)
    Y[sl] = X[sl]
    sl[axis] = slice(-(N-1)/2,None)
    Y[sl] = X[sl]
    y = ifft(Y,axis=axis)*(float(num)/float(Nx))

    if x.typecode() not in ['F','D']:
        y = y.real

    if t is None:
        return y
    else:
        new_t = arange(0,num)*(t[1]-t[0])* Nx / float(num) + t[0]
        return y, new_t

from scipy_base import expand_dims, unique, prod, sort, zeros, ones, \
     reshape, r_, any, c_, transpose, take, dot

import scipy.linalg as linalg
def detrend(data, axis=-1, type='linear', bp=0):
    """Remove linear trend along axis from data.

    If type is 'constant' then remove mean only.

    If bp is given, then it is a sequence of points at which to
       break a piecewise-linear fit to the data.
    """
    if type not in ['linear','l','constant','c']:
        raise ValueError, "Trend type must be linear or constant"
    data = asarray(data)
    dtype = data.typecode()
    if dtype not in 'dfDF':
        dtype = 'd'
    if type in ['constant','c']:
        ret = data - expand_dims(mean(data,axis),axis)
        return ret
    else:
        dshape = data.shape
        N = dshape[axis]
        bp = sort(unique(r_[0,bp,N]))
        if any(bp > N):
            raise ValueError, "Breakpoints must be less than length of data along given axis."
        Nreg = len(bp) - 1
        # Restructure data so that axis is along first dimension and
        #  all other dimensions are collapsed into second dimension
        rnk = len(dshape)
        if axis < 0: axis = axis + rnk
        newdims = r_[axis,0:axis,axis+1:rnk]
        newdata = reshape(transpose(data,tuple(newdims)),(N,prod(dshape)/N))
        newdata = newdata.copy()  # make sure we have a copy
        if newdata.typecode() not in 'dfDF':
            newdata = newdata.astype(dtype)
        # Find leastsq fit and remove it for each piece
        for m in range(Nreg):
            Npts = bp[m+1] - bp[m]
            A = ones((Npts,2),dtype)
            A[:,0] = cast[dtype](arange(1,Npts+1)*1.0/Npts)
            sl = slice(bp[m],bp[m+1])
            coef,resids,rank,s = linalg.lstsq(A,newdata[sl])
            newdata[sl] = newdata[sl] - dot(A,coef)
        # Put data back in original shape.
        tdshape = take(dshape,newdims)
        ret = reshape(newdata,tdshape)
        vals = range(1,rnk)
        olddims = vals[:axis] + [0] + vals[axis:]
        ret = transpose(ret,tuple(olddims))
        return ret




#
# Author: Pearu Peterson, March 2002
#

__all__ = ['get_blas_funcs']

import string

# The following ensures that possibly missing flavor (C or Fortran) is
# replaced with the available one. If none is available, exception
# is raised at the first attempt to use the resources.

import fblas
import cblas
if hasattr(cblas,'empty_module'):
    cblas = fblas
elif hasattr(fblas,'empty_module'):
    fblas = cblas

_type_conv = {'f':'s', 'd':'d', 'F':'c', 'D':'z'} # 'd' will be default for 'i',..
_inv_type_conv = {'s':'f','d':'d','c':'F','z':'D'}

def get_blas_funcs(names,arrays=(),debug=0):
    """Return available BLAS function objects with names.
    arrays are used to determine the optimal prefix of
    BLAS routines."""
    ordering = []
    for i in range(len(arrays)):
        t = arrays[i].typecode()
        if not _type_conv.has_key(t): t = 'd'
        ordering.append((t,i))
    if ordering:
        ordering.sort()
        required_prefix = _type_conv[ordering[0][0]]
    else:
        required_prefix = 'd'
    typecode = _inv_type_conv[required_prefix]
    # Default lookup:
    if ordering and fblas.has_column_major_storage(arrays[ordering[0][1]]):
        # prefer Fortran code for leading array with column major order
        m1,m2 = fblas,cblas
    else:
        # in all other cases, C code is preferred
        m1,m2 = cblas,fblas
    funcs = []
    for name in names:
        if name=='ger' and typecode in 'FD':
            name = 'gerc'
        func_name = required_prefix + name
        func = getattr(m1,func_name,None)
        if func is None:
            func = getattr(m2,func_name)
            func.module_name = string.split(m2.__name__,'.')[-1]
        else:
            func.module_name = string.split(m1.__name__,'.')[-1]
        func.prefix = required_prefix
        func.typecode = typecode
        funcs.append(func)
    return tuple(funcs)



# This file is executed by __init__.py and ppimport hooks.
"""
Discrete Fourier Transform algorithms
=====================================

Fast Fourier Transforms:

  fft       --- FFT of arbitrary type periodic sequences
  ifft      --- Inverse of fft
  fftn      --- Multi-dimensional FFT
  ifftn     --- Inverse of fftn
  fft2      --- Two-dimensional FFT
  ifft2     --- Inverse of fft2
  rfft      --- FFT of real periodic sequences
  irfft     --- Inverse of rfft

Differential and pseudo-differential operators:

  diff      --- Differentiation and integration of periodic sequences
  tilbert   --- Tilbert transform:         cs_diff(x,h,h)
  itilbert  --- Inverse Tilbert transform: sc_diff(x,h,h)
  hilbert   --- Hilbert transform:         cs_diff(x,inf,inf)
  ihilbert  --- Inverse Hilbert transform: sc_diff(x,inf,inf)
  cs_diff   --- cosh/sinh pseudo-derivative of periodic sequences
  sc_diff   --- sinh/cosh pseudo-derivative of periodic sequences
  ss_diff   --- sinh/sinh pseudo-derivative of periodic sequences
  cc_diff   --- cosh/cosh pseudo-derivative of periodic sequences
  shift     --- Shift periodic sequences

Helper functions:

  fftshift  --- Shift zero-frequency component to center of spectrum
  ifftshift --- Inverse of freqshift
  dftfreq   --- DFT sample frequencies
  rfftfreq  --- DFT sample frequencies (specific to rfft,irfft)

Extension modules:

  fftpack   --- Provides functions zfft, drfft, zrfft, zfftnd,
                destroy_*_cache
  convolve  --- Provides functions convolve, convolve_z,
                init_convolution_kernel, destroy_convolve_cache
"""

__all__ = ['fft','ifft','fftn','ifftn','rfft','irfft',
           'fft2','ifft2',
           'diff',
           'tilbert','itilbert','hilbert','ihilbert',
           'sc_diff','cs_diff','cc_diff','ss_diff',
           'shift',
           'fftshift','ifftshift',
           'fftfreq','rfftfreq'
           ]

__doc_title__ = __doc__.lstrip().split('\n',1)[0]

postpone_import = 1

global_symbols = ['fft','fftn','fft2','ifft','ifft2','ifftn',
                  'fftshift','ifftshift','fftfreq']



""" Class for interpolating values

    !! Need to find argument for keeping initialize.  If it isn't
    !! found, get rid of it!
"""

__all__ = ['interp1d']

from scipy_base import *
from scipy_base.fastumath import *

import fitpack

# The following are cluges to fix brain-deadness of take and
# sometrue when dealing with 0 dimensional arrays.
# Shouldn't they go to scipy_base??

_take = take
def take(a,indices,axis=0):    
    x = asarray(a); y = asarray(indices)
    if shape(x) == (): x = x.flat
    if shape(y) == (): y = y.flat
    return _take(x,y,axis)

_sometrue = sometrue
def sometrue(a,axis=0):    
    x = asarray(a)
    if shape(x) == (): x = x.flat
    return _sometrue(x)

def reduce_sometrue(a):
    all = a
    while len(shape(all)) > 1:    
        all = sometrue(all)
    return all

class interp2d:
    def __init__(self,x,y,z,kind='linear',
                 copy=1,bounds_error=0,fill_value=None):
        """
        Input:
          x,y  - 1-d arrays defining 2-d grid (or 2-d meshgrid arrays)
          z    - 2-d array of grid values
          kind - interpolation type ('nearest', 'linear', 'cubic', 'spline')
          copy - if true then data is copied into class, otherwise only a
                   reference is held.
          bounds_error - if true, then when out_of_bounds occurs, an error is
                          raised otherwise, the output is filled with
                          fill_value.
          fill_value - if None, then NaN, otherwise the value to fill in
                        outside defined region.
        """
        self.x = atleast_1d(x).copy()
        self.y = atleast_1d(y).copy()
        if rank(self.x) > 2 or rank(self.y) > 2:
            raise ValueError, "One of the input arrays is not 1-d or 2-d."
        if rank(self.x) == 2:
            self.x = self.x[:,0]
        if rank(self.y) == 2:
            self.y = self.y[0]
        self.z = array(z,copy=1)
        if rank(z) != 2:
            raise ValueError, "Grid values is not a 2-d array."


        

    def __call__(self,x,y,dx=0,dy=0):
        """
        Input:
          x,y   - 1-d arrays defining points to interpolate.
          dx,dy - order of partial derivatives in x and y, respectively.
                  0<=dx<kx, 0<=dy<ky
        Output:
          z     - 2-d array of interpolated values
        """
        x = atleast_1d(x)
        y = atleast_1d(y)
        z,ier=fitpack._fitpack._bispev(*(self.tck+[x,y,dx,dy]))
        if ier==10: raise ValueError,"Invalid input data"
        if ier: raise TypeError,"An error occurred"
        z.shape=len(x),len(y)
        z = transpose(z)
        if len(z)==1: z = z[0]
        return array(z)

class interp1d:
    interp_axis = -1 # used to set which is default interpolation
                     # axis.  DO NOT CHANGE OR CODE WILL BREAK.
                     
    def __init__(self,x,y,kind='linear',axis = -1,
                 copy = 1,bounds_error=1, fill_value=None):
        """Initialize a 1d linear interpolation class

        Description:
          x and y are arrays of values used to approximate some function f:
            y = f(x)
          This class returns a function whose call method uses linear
          interpolation to find the value of new points.

        Inputs:
            x -- a 1d array of monotonically increasing real values.
                 x cannot include duplicate values. (otherwise f is
                 overspecified)
            y -- an nd array of real values.  y's length along the
                 interpolation axis must be equal to the length
                 of x.
            kind -- specify the kind of interpolation: 'nearest', 'linear',
                    'cubic', or 'spline'
            axis -- specifies the axis of y along which to 
                    interpolate. Interpolation defaults to the last
                    axis of y.  (default: -1)
            copy -- If 1, the class makes internal copies of x and y.
                    If 0, references to x and y are used. The default 
                    is to copy. (default: 1)
            bounds_error -- If 1, an error is thrown any time interpolation
                            is attempted on a value outside of the range
                            of x (where extrapolation is necessary).
                            If 0, out of bounds values are assigned the
                            NaN (#INF) value.  By default, an error is
                            raised, although this is prone to change.
                            (default: 1)
        """      
        self.axis = axis
        self.copy = copy
        self.bounds_error = bounds_error
        if fill_value is None:
            self.fill_value = array(0.0) / array(0.0)
        else:
            self.fill_value = fill_value

        if kind != 'linear':
            raise NotImplementedError, "Only linear supported for now. Use fitpack routines for other types."
        
        # Check that both x and y are at least 1 dimensional.    
        if len(shape(x)) == 0 or len(shape(y)) == 0:
            raise ValueError, "x and y arrays must have at least one dimension."  
        # make a "view" of the y array that is rotated to the
        # interpolation axis.  
        oriented_x = x
        oriented_y = swapaxes(y,self.interp_axis,axis)            
        interp_axis = self.interp_axis        
        len_x,len_y = shape(oriented_x)[interp_axis], shape(oriented_y)[interp_axis]
        if len_x != len_y:
            raise ValueError, "x and y arrays must be equal in length along "\
                              "interpolation axis."
        if len_x < 2 or len_y < 2:
            raise ValueError, "x and y arrays must have more than 1 entry"            
        self.x = array(oriented_x,copy=self.copy)
        self.y = array(oriented_y,copy=self.copy)       
        
    def __call__(self,x_new):
        """Find linearly interpolated y_new = <name>(x_new).

        Inputs:        
          x_new -- New independent variables.

        Outputs:
          y_new -- Linearly interpolated values corresponding to x_new.
        """
        # 1. Handle values in x_new that are outside of x.  Throw error,
        #    or return a list of mask array indicating the outofbounds values.
        #    The behavior is set by the bounds_error variable.
        x_new = atleast_1d(x_new)
        out_of_bounds = self._check_bounds(x_new)
        # 2. Find where in the orignal data, the values to interpolate
        #    would be inserted.  
        #    Note: If x_new[n] = x[m], then m is returned by searchsorted.
        x_new_indices = searchsorted(self.x,x_new)
        # 3. Clip x_new_indices so that they are within the range of 
        #    self.x indices and at least 1.  Removes mis-interpolation
        #    of x_new[n] = x[0] 
        x_new_indices = clip(x_new_indices,1,len(self.x)-1).astype(Int)
        # 4. Calculate the slope of regions that each x_new value falls in.
        lo = x_new_indices - 1; hi = x_new_indices        
        
        # !! take() should default to the last axis (IMHO) and remove
        # !! the extra argument.
        x_lo = take(self.x,lo,axis=self.interp_axis)
        x_hi = take(self.x,hi,axis=self.interp_axis);
        y_lo = take(self.y,lo,axis=self.interp_axis)
        y_hi = take(self.y,hi,axis=self.interp_axis);
        slope = (y_hi-y_lo)/(x_hi-x_lo)
        # 5. Calculate the actual value for each entry in x_new.
        y_new = slope*(x_new-x_lo) + y_lo 
        # 6. Fill any values that were out of bounds with NaN
        # !! Need to think about how to do this efficiently for 
        # !! mutli-dimensional Cases.
        yshape = y_new.shape
        y_new = y_new.flat
        new_shape = list(yshape)
        new_shape[self.interp_axis] = 1
        sec_shape = [1]*len(new_shape)
        sec_shape[self.interp_axis] = len(out_of_bounds)
        out_of_bounds.shape = sec_shape
        new_out = ones(new_shape)*out_of_bounds
        putmask(y_new, new_out.flat, self.fill_value)
        y_new.shape = yshape      
        # Rotate the values of y_new back so that they coorespond to the
        # correct x_new values.
        result = swapaxes(y_new,self.interp_axis,self.axis)
        try:
            len(x_new)
            return result
        except TypeError:
            return result[0]
        return result
    
    def _check_bounds(self,x_new):
        # If self.bounds_error = 1, we raise an error if any x_new values
        # fall outside the range of x.  Otherwise, we return an array indicating
        # which values are outside the boundary region.  
        # !! Needs some work for multi-dimensional x !!
        below_bounds = less(x_new,self.x[0])
        above_bounds = greater(x_new,self.x[-1])
        #  Note: sometrue has been redefined to handle length 0 arrays
        # !! Could provide more information about which values are out of bounds
        if self.bounds_error and sometrue(below_bounds):
            raise ValueError, " A value in x_new is below the"\
                              " interpolation range."
        if self.bounds_error and sometrue(above_bounds):
            raise ValueError, " A value in x_new is above the"\
                              " interpolation range."
        # !! Should we emit a warning if some values are out of bounds.
        # !! matlab does not.
        out_of_bounds = logical_or(below_bounds,above_bounds)
        return out_of_bounds
       
    def model_error(self,x_new,y_new):
        # How well do x_new,yy points fit the model?
        # Return an array of error values.
        pass

    
#assumes module test_xxx is in python path
#def test():
#    test_module = 'test_' + __name__ # __name__ is name of this module
#    test_string = 'import %s;reload(%s);%s.test()' % ((test_module,)*3)
#    exec(test_string)

#if __name__ == '__main__':
#    test()



#
# cow - Cluster Of Workstations
#

from info_cow import __doc__

import sync_cluster
from cow import *

Mrange = range(1,17)    
import cow # cow stands for "Cluster Of Workstations"
server_list = []
server_list.append((sync_cluster.host,10000))

local = cow.machine_cluster(server_list)




from scipy_test.testing import *

set_package_path()
from optimize import cobyla as co
restore_path()

class test_cobyla(ScipyTestCase):
    def check_simple(self, level=1):

        function = lambda x: x[0]**2 + abs(x[1])**3
	con1 = lambda x: x[0]**2 + x[1]**2 - 25
	con2 = lambda x: -con1(x)

        x = co.fmin_cobyla(function, [4.95,0.66], [con1, con2], rhobeg=1, 
			   rhoend=1e-5, iprint=0, maxfun=100)
        print 'Result:',x,'(exact result = 4.955356249106168, 0.666666666666666)'

if __name__ == "__main__":
    ScipyTest().run()



""" Test functions for limits module.

    Currently empty -- not sure how to test these values
    and routines as they are machine dependent.  Suggestions?
"""

import unittest

import sys
from scipy_test.testing import *
set_package_path()
import scipy_base;reload(scipy_base)
from scipy_base import *
del sys.path[0]

##################################################
### Test for sum

class test_float(unittest.TestCase):
    def check_nothing(self):
        pass

class test_double(unittest.TestCase):
    def check_nothing(self):
        pass

if __name__ == "__main__":
    ScipyTest('scipy_base.limits').run()




""" Auto test tools for SciPy

    Do not run this as root!  If you enter something
    like /usr as your test directory, it'll delete
    /usr/bin, usr/lib, etc.  So don't do it!!!
    
    
    Author: Eric Jones (eric@enthought.com)
"""
from distutils import file_util
from distutils import dir_util
from distutils.errors import DistutilsFileError
#import tarfile
import sys, os, stat, time
import gzip
import tempfile, cStringIO
import urllib
import logging

if sys.platform == 'cygwin':
    local_repository = "/cygdrive/i/tarballs"
elif sys.platform == 'win32':    
    local_repository = "i:\tarballs"
else:
    local_repository = "/home/shared/tarballs"

local_mail_server = "enthought.com"

python_ftp_url = "ftp://ftp.python.org/pub/python"
numeric_url = "http://prdownloads.sourceforge.net/numpy"
f2py_url = "http://cens.ioc.ee/projects/f2py2e/2.x"
scipy_url = "ftp://www.scipy.org/pub"
blas_url = "http://www.netlib.org/blas"
lapack_url = "http://www.netlib.org/lapack"
#atlas_url = "http://prdownloads.sourceforge.net/math-atlas"
atlas_url = "http://www.scipy.org/Members/eric"


#-----------------------------------------------------------------------------
# Generic installation class. 
# built to handle downloading/untarring/building/installing arbitrary software
#-----------------------------------------------------------------------------

class package_installation:    
    def __init__(self,version='', dst_dir = '.',
                 logger = None, python_exe='python'):
        #---------------------------------------------------------------------
        # These should be defined in sub-class before calling this
        # constructor
        #---------------------------------------------------------------------
        # 
        #self.package_url -- The name of the url where tarball can be found.
        #self.package_base_name -- The base name of the source tarball.
        #self.package_dir_name -- Top level directory of unpacked tarball
        #self.tarball_suffix -- usually tar.gz or .tgz
        #self.build_type -- 'make' or 'setup' for makefile or python setup file
        
        # Version of the software package.
        self.version = version

        # Only used by packages built with setup.py
        self.python_exe = python_exe
        
        # Directory where package is unpacked/built/installed
        self.dst_dir = os.path.abspath(dst_dir)        
        
        if not logger:
            self.logger = logging
        else:
            self.logger = logger    

        # make sure the destination exists
        make_dir(self.dst_dir,logger=self.logger)

        # Construct any derived names built from the above names.
        self.init_names()
        
    def init_names(self):            
        self.package_dir = os.path.join(self.dst_dir,self.package_dir_name)
        self.tarball = self.package_base_name + '.' + self.tarball_suffix

    def get_source(self):
        """ Grab the source tarball from a repository.
        
            Try a local repository first.  If the file isn't found,
            grab it from an ftp site.
        """
        local_found = 0
        if self.local_source_up_to_date():
            try:
                self.get_source_local()
                local_found = 1                
            except DistutilsFileError:
                pass
        
        if not local_found:
            self.get_source_ftp()
                    
    def local_source_up_to_date(self):
        """ Hook to test whether a file found in the repository is current
        """
        return 1
        
    def get_source_local(self):
        """ Grab the requested tarball from a local repository of source
            tarballs.  If it doesn't exist, an error is raised.
        """
        file = os.path.join(local_repository,self.tarball)        
        dst_file = os.path.join(self.dst_dir,self.tarball)
        self.logger.info("Searching local repository for %s" % file)
        try:
            copy_file(file,dst_file,self.logger)
        except DistutilsFileError, msg:
            self.logger.info("Not found:",msg)
            raise
        
    def get_source_ftp(self):
        """ Grab requested tarball from a ftp site specified as a url.           
        """
        url = '/'.join([self.package_url,self.tarball])
     
        self.logger.info('Opening: %s' % url)
        f = urllib.urlopen(url)
        self.logger.info('Downloading: this may take a while')
        contents = f.read(-1)
        f.close()
        self.logger.info('Finished download (size=%d)' % len(contents))
     
        output_file = os.path.join(self.dst_dir,self.tarball)
        write_file(output_file,contents,self.logger)

        # Put file in local repository so we don't have to download it again.
        self.logger.info("Caching file in repository" )
        src_file = output_file
        repos_file = os.path.join(local_repository,self.tarball)        
        copy_file(src_file,repos_file,self.logger)

    def unpack_source(self,sub_dir = None):
        """ equivalent to 'tar -xzvf file' in the given sub_dir
        """       
        tarfile = os.path.join(self.dst_dir,self.tarball)
        old_dir = None
        
        # copy and move into sub directory if it is specified.
        if sub_dir:
            dst_dir = os.path.join(self.dst_dir,sub_dir)
            dst_file = os.path.join(dst_dir,self.tarball)
            copy_file(tarfile,dst_file)
            change_dir(dst_dir,self.logger)
        try:
            try:
                # occasionally the tarball is not zipped, try this first.
                untar_file(self.tarball,self.dst_dir,
                           self.logger,silent_failure=1)
            except:
                # otherwise, handle the fact that it is zipped        
                dst = os.path.join(self.dst_dir,'tmp.tar')        
                decompress_file(tarfile,dst,self.logger)                
                untar_file(dst,self.dst_dir,self.logger)
                remove_file(dst,self.logger)
        finally:
            if old_dir:
                unchange_dir(self.logger)

    #def auto_configure(self):
    #    cmd = os.path.join('.','configure')
    #    try:
    #        text = run_command(cmd,self.package_dir,self.logger,log_output=0)
    #    except ValueError, e:
    #        status, text = e
    #        self.logger.exception('Configuration Error:\n'+text)
    def auto_configure(self):
        cmd = os.path.join('.','configure')
        text = run_command(cmd,self.package_dir,self.logger)
        
    def build_with_make(self):
        cmd = 'make'
        text = run_command(cmd,self.package_dir,self.logger)
        
    def install_with_make(self, prefix = None):
        if prefix is None:
            prefix = os.path.abspath(self.dst_dir)
        cmd = 'make install prefix=%s' % prefix
        text = run_command(cmd,self.package_dir,self.logger)
        
    def python_setup(self):
        cmd = self.python_exe + ' setup.py install'
        text = run_command(cmd,self.package_dir,self.logger)
        
    def _make(self,**kw):
        """ This generally needs to be overrridden in the derived class,
            but this will suffice for the standard configure/make process.            
        """
        self.logger.info("### Begin Configure: %s" % self.package_base_name)
        self.auto_configure()
        self.logger.info("### Finished Configure: %s" % self.package_base_name)
        self.logger.info("### Begin Build: %s" % self.package_base_name)
        self.build_with_make()
        self.logger.info("### Finished Build: %s" % self.package_base_name)
        self.logger.info("### Begin Install: %s" % self.package_base_name)
        self.install_with_make()
        self.logger.info("### Finished Install: %s" % self.package_base_name)

    def install(self):
        self.logger.info('####### Building:    %s' % self.package_base_name)
        self.logger.info('        Version:     %s' % self.version)
        self.logger.info('        Url:         %s' % self.package_url)
        self.logger.info('        Install dir: %s' % self.dst_dir)
        self.logger.info('        Package dir: %s' % self.package_dir)
        self.logger.info('        Suffix:      %s' % self.tarball_suffix)
        self.logger.info('        Build type:  %s' % self.build_type)

        self.logger.info("### Begin Get Source: %s" % self.package_base_name)
        self.get_source()
        self.unpack_source()
        self.logger.info("### Finished Get Source: %s" % self.package_base_name)

        if self.build_type == 'setup':
            self.python_setup()
        else:    
            self._make()
        self.logger.info('####### Finished Building: %s' % self.package_base_name)            
            
#-----------------------------------------------------------------------------
# Installation class for Python itself.
#-----------------------------------------------------------------------------
        
class python_installation(package_installation):
    
    def __init__(self,version='', dst_dir = '.',logger=None,python_exe='python'):
        
        # Specialization for Python.        
        self.package_base_name = 'Python-'+version
        self.package_dir_name = self.package_base_name
        self.package_url = '/'.join([python_ftp_url,version])
        self.tarball_suffix = 'tgz'
        self.build_type = 'make'
        
        package_installation.__init__(self,version,dst_dir,logger,python_exe)

    def write_install_config(self):    
        """ Make doesn't seem to install scripts in the correct places.
        
            Writing this to the python directory will solve the problem.
            [install_script]
            install-dir=<directory_name> 
        """
        self.logger.info('### Writing Install Script Hack')
        text = "[install_scripts]\n"\
               "install-dir='%s'" % os.path.join(self.dst_dir,'bin')
        file = os.path.join(self.package_dir,'setup.cfg')               
        write_file(file,text,self.logger,mode='w')
        self.logger.info('### Finished writing Install Script Hack')

    def install_with_make(self):
        """ Scripts were failing to install correctly, so a setuo.cfg
            file is written to force installation in the correct place.
        """        
        self.write_install_config()
        package_installation.install_with_make(self)

    def get_exe_name(self):
        pyname = os.path.join('.','python')
        cmd = pyname + """ -c "import sys;print '%d.%d' % sys.version_info[:2]" """
        text = run_command(cmd,self.package_dir,self.logger)
        exe = os.path.join(self.dst_dir,'bin','python'+text)
        return exe

#-----------------------------------------------------------------------------
# Installation class for Blas.
#-----------------------------------------------------------------------------

class blas_installation(package_installation):
    
    def __init__(self,version='', dst_dir = '.',logger=None,python_exe='python'):
        
        # Specialization for for "slow" blas
        self.package_base_name = 'blas'
        self.package_dir_name = 'BLAS'
        self.package_url = blas_url
        self.tarball_suffix = 'tgz'
        self.build_type = 'make'
                
        self.platform = 'LINUX'
        package_installation.__init__(self,version,dst_dir,logger,python_exe)

    def unpack_source(self,subdir=None):
        """ Dag.  blas.tgz doesn't have directory information -- its
            just a tar ball of fortran source code.  untar it in the
            BLAS directory
        """
        package_installation.unpack_source(self,self.package_dir_name)
            
    def auto_configure(self):
        # nothing to do.
        pass
    def build_with_make(self, **kw):
        libname = 'blas_LINUX.a'
        cmd = 'g77 -funroll-all-loops -fno-f2c -O3 -c *.f;ar -cru %s' % libname
        text = run_command(cmd,self.package_dir,self.logger)
        
    def install_with_make(self, **kw):
        # not really using make -- we'll just copy the file over.        
        src_file = os.path.join(self.package_dir,'blas_%s.a' % self.platform)
        dst_file = os.path.join(self.dst_dir,'lib','libblas.a')
        self.logger.info("Installing blas")
        copy_file(src_file,dst_file,self.logger)
        
#-----------------------------------------------------------------------------
# Installation class for Lapack.
#-----------------------------------------------------------------------------

class lapack_installation(package_installation):
    
    def __init__(self,version='', dst_dir = '.',logger=None,python_exe='python'):
        
        # Specialization for Lapack 3.0 + updates        
        self.package_base_name = 'lapack'
        self.package_dir_name = 'LAPACK'
        self.package_url = lapack_url
        self.tarball_suffix = 'tgz'
        self.build_type = 'make'
        
        self.platform = 'LINUX'
        package_installation.__init__(self,version,dst_dir,logger,python_exe)

    def auto_configure(self):
        # perhaps this should actually override auto_conifgure
        # before make, we need to copy the appropriate setup file in.
        # should work anywhere g77 works...
        make_inc = 'make.inc.' + self.platform
        src_file = os.path.join(self.package_dir,'INSTALL',make_inc)
        dst_file = os.path.join(self.package_dir,'make.inc')
        copy_file(src_file,dst_file,self.logger)
        
    def build_with_make(self, **kw):
        cmd = 'make install lapacklib'
        text = run_command(cmd,self.package_dir,self.logger)
        
    def install_with_make(self, **kw):
        # not really using make -- we'll just copy the file over.
        src_file = os.path.join(self.package_dir,'lapack_%s.a' % self.platform)
        dst_file = os.path.join(self.dst_dir,'lib','liblapack.a')        
        copy_file(src_file,dst_file,self.logger)

#-----------------------------------------------------------------------------
# Installation class for Numeric
#-----------------------------------------------------------------------------

class numeric_installation(package_installation):
    
    def __init__(self,version='', dst_dir = '.',logger=None,python_exe='python'):
        
        self.package_base_name = 'Numeric-'+version
        self.package_dir_name = self.package_base_name
        self.package_url = numeric_url
        self.tarball_suffix = 'tar.gz'
        self.build_type = 'setup'        

        package_installation.__init__(self,version,dst_dir,logger,python_exe)


#-----------------------------------------------------------------------------
# Installation class for f2py
#-----------------------------------------------------------------------------

class f2py_installation(package_installation):
    
    def __init__(self,version='', dst_dir = '.',logger=None,python_exe='python'):
        
        # Typical file format: F2PY-2.13.175-1250.tar.gz
        self.package_base_name = 'F2PY-'+version
        self.package_dir_name = self.package_base_name
        self.package_url = f2py_url
        self.tarball_suffix = 'tar.gz'
        self.build_type = 'setup'        
                
        package_installation.__init__(self,version,dst_dir,logger,python_exe)


#-----------------------------------------------------------------------------
# Installation class for Atlas.
# This is a binary install *NOT* a source install.
# The source install is a pain to automate.
#-----------------------------------------------------------------------------

class atlas_installation(package_installation):
    
    def __init__(self,version='', dst_dir = '.',logger=None,python_exe='python'):
        
        #self.package_base_name = 'atlas' + version
        #self.package_dir_name = 'ATLAS'
        self.package_base_name = 'atlas-RH7.1-PIII'
        self.package_dir_name = 'atlas'
        self.package_url = atlas_url
        self.tarball_suffix = 'tgz'
        self.build_type = 'make'        
        
        package_installation.__init__(self,version,dst_dir,logger,python_exe)

    def auto_configure(self,**kw):
        pass
    def build_with_make(self,**kw):
        pass
    def install_with_make(self, **kw):
        # just copy the tree over.
        dst = os.path.join(self.dst_dir,'lib','atlas')
        self.logger.info("Installing Atlas")
        copy_tree(self.package_dir,dst,self.logger)

#-----------------------------------------------------------------------------
# Installation class for scipy
#-----------------------------------------------------------------------------

class scipy_installation(package_installation):
    
    def __init__(self,version='', dst_dir = '.',logger=None,python_exe='python'):
        
        self.package_base_name = 'scipy_snapshot'
        self.package_dir_name = 'scipy'
        self.package_url = scipy_url
        self.tarball_suffix = 'tgz'
        self.build_type = 'setup'
        
        package_installation.__init__(self,version,dst_dir,logger,python_exe)
                    
    def local_source_up_to_date(self):
        """ Hook to test whether a file found in the repository is current
        """
        file = os.path.join(local_repository,self.tarball)
        up_to_date = 0
        try:
            file_time = os.stat(file)[stat.ST_MTIME]        
            fyear,fmonth,fday = time.localtime(file_time)[:3]
            year,month,day = time.localtime()[:3]
            if fyear == year and fmonth == month and fday == day:
                up_to_date = 1
                self.logger.info("Repository file up to date: %s" % file)
        except OSError, msg:
            pass
        return up_to_date
                
#-----------------------------------------------------------------------------
# Utilities
#-----------------------------------------------------------------------------


#if os.name == 'nt':
#    def exec_command(command):
#        """ not sure how to get exit status on nt. """
#        in_pipe,out_pipe = os.popen4(command)
#        in_pipe.close()
#        text = out_pipe.read()
#        return 0, text
#else:
#    import commands
#    exec_command = commands.getstatusoutput
   
# This may not work on Win98... The above stuff was to handle these machines.
import commands
exec_command = commands.getstatusoutput

def copy_file(src,dst,logger=None):
    if not logger:
        logger = logging
    logger.info("Copying %s->%s" % (src,dst))        
    try:
        file_util.copy_file(src,dst)
    except Exception, e:     
        logger.exception("Copy Failed")        
        raise

def copy_tree(src,dst,logger=None):
    if not logger:
        logger = logging
    logger.info("Copying directory tree %s->%s" % (src,dst))        
    try:
        dir_util.copy_tree(src,dst)
    except Exception, e:     
        logger.exception("Copy Failed")        
        raise

def remove_tree(directory,logger=None):
    if not logger:
        logger = logging
    logger.info("Removing directory tree %s" % directory)        
    try:
        dir_util.remove_tree(directory)
    except Exception, e:     
        logger.exception("Remove failed: %s" % e)        
        raise

def remove_file(file,logger=None):
    if not logger:
        logger = logging
    logger.info("Remove file %s" % file)        
    try:
        os.remove(file)
    except Exception, e:     
        logger.exception("Remove failed")        
        raise

def write_file(file,contents,logger=None,mode='wb'):
    if not logger:
        logger = logging
    logger.info('Write file: %s' % file)
    try:
        new_file = open(file,mode)
        new_file.write(contents)
        new_file.close()
    except Exception, e:     
        logger.exception("Write failed")        
        raise

def make_dir(name,logger=None):
    if not logger:
        logger = logging
    logger.info('Make directory: %s' % name)
    try:        
        dir_util.mkpath(os.path.abspath(name))
    except Exception, e:     
        logger.exception("Make Directory failed")        
        raise

# I know, I know...
old_dir = []

def change_dir(d, logger = None):
    if not logger:
        logger = logging
    global old_dir 
    cwd = os.getcwd()   
    old_dir.append(cwd)
    d = os.path.abspath(d)
    if d != old_dir[-1]:
        logger.info("Change directory: %s" % d)            
        try:
            os.chdir(d)
        except Exception, e:     
            logger.exception("Change directory failed")
            raise        
        #if d == '.':
        #    import sys,traceback
        #    f = sys._getframe()
        #    traceback.print_stack(f)

def unchange_dir(logger=None):
    if not logger:
        logger = logging            
    global old_dir
    try:
        cwd = os.getcwd()
        d = old_dir.pop(-1)            
        try:
            if d != cwd:
                logger.info("Change directory : %s" % d)
                os.chdir(d)
        except Exception, e:     
            logger.exception("Change directory failed")
            raise                    
    except IndexError:
        logger.exception("Change directory failed")
        
def decompress_file(src,dst,logger = None):
    if not logger:
        logger = logging
    logger.info("Upacking %s->%s" % (src,dst))
    try:
        f = gzip.open(src,'rb')
        contents = f.read(-1)
        f = open(dst, 'wb')
        f.write(contents)
    except Exception, e:     
        logger.exception("Unpack failed")
        raise        

    
def untar_file(file,dst_dir='.',logger = None,silent_failure = 0):    
    if not logger:
        logger = logging
    logger.info("Untarring file: %s" % (file))
    try:
        run_command('tar -xf ' + file,directory = dst_dir,
                    logger=logger, silent_failure = silent_failure)
    except Exception, e:
        if not silent_failure:     
            logger.exception("Untar failed")
        raise        

def unpack_file(file,logger = None):
    """ equivalent to 'tar -xzvf file'
    """
    dst = 'tmp.tar'
    decompress_file(file,dst,logger)                
    untar_file(dst.logger)
    remove_file(dst,logger)        


def run_command(cmd,directory='.',logger=None,silent_failure = 0):
    if not logger:
        logger = logging
    change_dir(directory,logger)    
    try:        
        msg = 'Running: %s' % cmd
        logger.info(msg)    
        status,text = exec_command(cmd)
        if status and silent_failure:
            msg = '(failed silently)'
            logger.info(msg)    
        if status and text and not silent_failure:
            logger.error('Command Failed (status=%d)\n'% status +text)
    finally:
        unchange_dir(logger)
    if status:
        raise ValueError, (status,text)
    return text            

def mail_report(from_addr,to_addr,subject,mail_server,
                build_log, test_results,info):
    
    msg = ''
    msg = msg + 'To: %s\n'   % to_addr
    msg = msg + 'Subject: %s\n' % subject
    msg = msg + '\r\n\r\n'

    for k,v in info.items():   
        msg = msg + '%s: %s\n' % (k,v)
    msg = msg + test_results + '\n'
    msg = msg + '-----------------------------\n' 
    msg = msg + '--------  BUILD LOG   -------\n' 
    msg = msg + '-----------------------------\n' 
    msg = msg + build_log
    print msg
    
    # mail results
    import smtplib 
    server = smtplib.SMTP(mail_server)    
    server.sendmail(from_addr, to_addr, msg)
    server.quit()
    

def full_scipy_build(build_dir = '.',
                     test_level = 10,
                     python_version  = '2.2.1',
                     numeric_version = '21.0',
                     f2py_version    = '2.13.175-1250',
                     atlas_version   = '3.3.14',
                     scipy_version   = 'snapshot'):
    
    # for now the atlas version is ignored.  Only the 
    # binaries for RH are supported at the moment.

    build_info = {'python_version' : python_version,
                  'test_level'     : test_level,
                  'numeric_version': numeric_version,
                  'f2py_version'   : f2py_version,
                  'atlas_version'  : atlas_version,
                  'scipy_version'  : scipy_version}
                    
    dst_dir = os.path.join(build_dir,sys.platform)

    logger = logging.Logger("SciPy Test")
    fmt = logging.Formatter(logging.BASIC_FORMAT)
    log_stream = cStringIO.StringIO()
    stream_handler = logging.StreamHandler(log_stream)
    stream_handler.setFormatter(fmt)
    logger.addHandler(stream_handler)
    # also write to stderr
    stderr = logging.StreamHandler()
    stderr.setFormatter(fmt)
    logger.addHandler(stderr)

    try:
        try:    
        
            # before doing anything, we need to wipe the 
            # /bin, /lib, /man, and /include directories
            # in dst_dir.  Don't run as root.            
            make_dir(dst_dir,logger=logger)            
            change_dir(dst_dir   , logger)
            for d in ['bin','lib','man','include']:
                try:            remove_tree(d, logger)
                except OSError: pass                
            unchange_dir(logger)
            
            python = python_installation(version=python_version,
                                         logger = logger,
                                         dst_dir = dst_dir)
            python.install()
            
            python_name = python.get_exe_name()
        
            numeric = numeric_installation(version=numeric_version,
                                           dst_dir = dst_dir,
                                           logger = logger,
                                           python_exe=python_name)
            numeric.install()
            
            f2py =  f2py_installation(version=f2py_version,
                                      logger = logger,
                                      dst_dir = dst_dir,
                                      python_exe=python_name)
            f2py.install()                                
        
            # download files don't have a version specified    
            #lapack =  lapack_installation(version='',
            #                              dst_dir = dst_dir
            #                              python_exe=python_name)
            #lapack.install()                                
        
            # download files don't have a version specified    
            #blas =  blas_installation(version='',
            #                          logger = logger,
            #                          dst_dir = dst_dir,
            #                          python_exe=python_name)
            #blas.install()                                
            
            # ATLAS
            atlas =  atlas_installation(version=atlas_version,
                                        logger = logger,
                                        dst_dir = dst_dir,
                                        python_exe=python_name)
            atlas.install()
            
            # version not currently used -- need to fix this.
            scipy =  scipy_installation(version=scipy_version,
                                        logger = logger,
                                        dst_dir = dst_dir,
                                        python_exe=python_name)
            scipy.install()                                
        
            # The change to tmp makes sure there isn't a scipy directory in 
            # the local scope.
            # All tests are run.
            logger.info('Beginning Test')
            cmd = python_name +' -c "import sys,scipy;suite=scipy.test(%d);"'\
                                % test_level
            test_results = run_command(cmd, logger=logger,
                                       directory = tempfile.gettempdir())
            build_info['results'] = 'test completed (check below for pass/fail)'
        except Exception, msg:
            test_results = ''
            build_info['results'] = 'build failed: %s' % msg
            logger.exception('Build failed')
    finally:    
        to_addr = "scipy-testlog@scipy.org"
        from_addr = "scipy-test@enthought.com"
        subject = '%s,py%s,num%s,scipy%s' % (sys.platform,python_version,
                                            numeric_version,scipy_version) 
        build_log = log_stream.getvalue()
        mail_report(from_addr,to_addr,subject,local_mail_server,
                    build_log,test_results,build_info)

if __name__ == '__main__':
    build_dir = '/tmp/scipy_test'
    level = 10

    full_scipy_build(build_dir = build_dir,
                     test_level = level,
                     python_version  = '2.2.1',
                     numeric_version = '21.0',
                     f2py_version    = '2.13.175-1250',
                     atlas_version   = '3.3.14',
                     scipy_version   = 'snapshot')

    # an older python
    full_scipy_build(build_dir = build_dir,
                     test_level = level,
                     python_version  = '2.1.3',
                     numeric_version = '21.0',
                     f2py_version    = '2.13.175-1250',
                     atlas_version   = '3.3.14',
                     scipy_version   = 'snapshot')

    # an older numeric
    full_scipy_build(build_dir = build_dir,
                     test_level = level,
                     python_version  = '2.1.3',
                     numeric_version = '20.3',
                     f2py_version    = '2.13.175-1250',
                     atlas_version   = '3.3.14',
                     scipy_version   = 'snapshot')

    # This fails because multiarray doesn't have 
    # arange defined.
    """
    full_scipy_build(build_dir = build_dir,
                     test_level = level,
                     python_version  = '2.1.3',
                     numeric_version = '20.0.0',
                     f2py_version    = '2.13.175-1250',
                     atlas_version   = '3.3.14',
                     scipy_version   = 'snapshot')

    full_scipy_build(build_dir = build_dir,
                     test_level = level,
                     python_version  = '2.1.3',
                     numeric_version = '19.0.0',
                     f2py_version    = '2.13.175-1250',
                     atlas_version   = '3.3.14',
                     scipy_version   = 'snapshot')

    full_scipy_build(build_dir = build_dir,
                     test_level = level,
                     python_version  = '2.1.3',
                     numeric_version = '18.4.1',
                     f2py_version    = '2.13.175-1250',
                     atlas_version   = '3.3.14',
                     scipy_version   = 'snapshot')
    """



#this just adds errorbar plotting capabilities
from Numeric import *
from scipy_base.scimath import *
from types import StringType
import pyPlot
from pyPlot import *

def squeeze(a):
	s = shape(a)
	dims = len(s) - sum(equal(1,s))
	new_s = compress(not_equal(1,s),s)
	return reshape(a,new_s)


class Plot(pyPlot.Plot):
	errorbar_lt = 1
	def errorbar(self,*data):
		self._init_plot()
		self._errorbar_plot(data)	

	def _errorbar_plot(self,data):
		remains = data
		counter = 0
		plotcmd = ''
		tmpf = []
		if not self.m_hold: self.errorbar_lt = 1
		while len(remains) > 0:
			group, remains = self._errorbar_get_plot_group(remains)
			if(len(group) > 0):
				try:
					cmd, file = self._errorbar_prepare_plot(group)
				except SizeMismatch, data:					
					print "Warning: ignoring plot group ", counter
					print "The arrays should have the same number of " + data[0]
					print 'array sizes: ', data[1],'and', data[2]
				else:
					plotcmd = plotcmd + cmd
					tmpf.append(file) 					
			counter = counter + 1
			
		# if the plot command was valid, and we aren't
		# adding to the current plot, delete all the old plot files	
		if(len(plotcmd) > 0):
			if(not self.m_hold):			
				self._delete_files()
			plotcmd = plotcmd[0:-2] # get rid of trailing comma
			if DEBUG == 1:
				print plotcmd
			if(self.m_hold == 1 and len(self.m_tmpfiles) > 0):
				self._send('replot ' + plotcmd)			
			else:				
				self._send('plot ' + plotcmd)			
			self.m_tmpfiles = self.m_tmpfiles + tmpf
		
	def _errorbar_prepare_plot(self, group):
		plotinfo = ''
		x = group[0]
		y = group[1]				
		e = group[2]
		ar_num = 3
		if(len(group) > 3):
			plotinfo = group[3]
		
#force 1D arrays to 2D columns
		if(is1D(x) ): x = col(x)
		if(is1D(y) ): y = col(y)
		if(is1D(e) ): e = col(e)
		xs = shape(x)
		ys = shape(y)	
		es = shape(e)
#test that x and y have compatible shapes
		#check that each array has the same number of rows
		if(xs[0] != ys[0] or xs[0] != es[0]):
			raise SizeMismatch, ('rows', xs, ys, es)
		#check that x.cols = y.cols
		#no error x has 1 column
		if(xs[1] > 1 and (xs[1] != ys[1] or xs[1] !=es[1])):
				raise SizeMismatch, ('cols', xs, ys, es)
		if(ys[1] != es[1]):
			raise SizeMismatch, ('cols', xs, ys, es)		
#now write the data to a temp file and then plot it			
		filename = tempfile.mktemp()
		f = open(filename, 'w')
		using = ()

		mn,dummy = self._find_min_max(y-e)
		dummy,mx = self._find_min_max(y+e)
		if(mn < self.m_rmin):
			self.m_rmin = mn
		if(mx > self.m_rmax):
			self.m_rmax = mx
		
		#x is effectively 1D
		if(xs[1] == 1):
			for i in range(0,xs[0] ):
				f.write( self._format( (x[i][0],) ))
				for j in range(0,ys[1] ):
					f.write( self._format( (y[i][j],e[i][j]) ))
				f.write('\n')			
			for j in range(0,ys[1] ):
				using = using + (' using 1:%d:%d' % (j+2,j+3),)
			#x is 2D					
		else:
			for i in range(0,xs[0] ):
				for j in range(0,xs[1] ):
					f.write( self._format( (x[i][j],y[i][j],e[i][j]) ) )
				f.write('\n')			
			for j in range(0,xs[1] ):
				using = using + (' using %d:%d:%d' % (j*2+1,j*2+2,j*2+3),)
		if plotinfo == '':
			#defualt to printing with lines	
			plotinfo = " notitle w lp  lt " +`self.errorbar_lt` + ' '
		cmd = ''
		# gnuplot doesn't seem to like backslashes
		fn = 	string.replace(filename,'\\','/')
		for us in using:
			cmd = cmd + ' "' + fn + '" ' + us + 'notitle w errorbars lt ' +`self.errorbar_lt` + ', '
			cmd = cmd + ' "' + fn + '" ' + us + plotinfo + ', '
		self.errorbar_lt = (self.errorbar_lt + 1) % 15
		return cmd, filename	
		
	def _errorbar_get_plot_group(self,data):
		group = ()
		remains = data
		state = 0
		finished = 0
		while(len(remains) > 0 and not finished):
			el = remains[0]
			if(state == 0):
				el = asarray(el)
				state = 1 
			elif(state == 1):
				el = asarray(el)
				state = 2
			elif(state == 2):
				if(type(el) == StringType):
					finished = 1
				else:
					el = asarray(el)
					state = 3
			elif(state == 3):
				finished = 1
				if(type(el) != StringType):
					break					
			group = group + (el,)
			remains = remains[1:]		
		return group, remains	



"""
"""

from plot_utility import *
from wxPython     import wx
from colormap     import colormap_map

#-----------------------------------------------------------------------#
#------- Attribute list/functions for objects in wxPython --------------#
#-----------------------------------------------------------------------#

# Actually the wxWindows color dialog is used to set most
# color properties, so this can be ignored.  The list of supported
# name colors is much longer than this

colors = ['black','blue','green','red','yellow','cyan','brown','orange',
          'purple','grey','white','light grey']
line_widths = [1,2,3,4,5]
tick_length = [0,1,2,3,4,5,6,7,8,9,10]

line_style_map = {'solid':wx.wxSOLID,'dot dash': wx.wxDOT_DASH,
                 'dash':wx.wxSHORT_DASH,'dot': wx.wxDOT,
                 'long dash':wx.wxLONG_DASH,'transparent':wx.wxTRANSPARENT}   
line_styles = line_style_map.keys()      

fill_style_map = {'solid':wx.wxSOLID,'stipple': wx.wxSTIPPLE,
                 'back hatch':wx.wxBDIAGONAL_HATCH,
                 'diag hatch': wx.wxCROSSDIAG_HATCH,
                 'forward hatch':wx.wxFDIAGONAL_HATCH,
                 'cross hatch': wx.wxCROSS_HATCH,
                 'horz hatch': wx.wxHORIZONTAL_HATCH,
                 'vert hatch': wx.wxVERTICAL_HATCH,
                 'transparent':wx.wxTRANSPARENT
                 }                 
fill_styles = fill_style_map.keys()

image_type_map = { 'jpg': wx.wxBITMAP_TYPE_JPEG,
                   'jpeg': wx.wxBITMAP_TYPE_JPEG,
                   'png': wx.wxBITMAP_TYPE_PNG,
                   'pcx': wx.wxBITMAP_TYPE_PCX,
                   'tif': wx.wxBITMAP_TYPE_TIF,
                   'tiff': wx.wxBITMAP_TYPE_TIF, }
    
def get_color(in_color):
    """ Convert a color name or  rgb sequence to a wxColour
    """
    if type(in_color) == type(''):
        color = wx.wxNamedColour(in_color) 
    else: 
        r,g,b = in_color
        ##color = wx.wxColour(r,g,b) # mod by GAP 26092003
        color = wx.wxColour(int(r),int(g),int(b)) 
    return color

def default_font():
    """ The default font for text objects.
        This is a method instead of a global variable so that
        each text object will get its own font object that can
        change indepedently of all others.
    """
    return wx.wxFont(12, wx.wxSWISS, wx.wxNORMAL, 
                     wx.wxNORMAL,wx.false,"Arial")                


#-----------------------------------------------------------------------#
#------------------------- Drawing utilities ---------------------------#
#-----------------------------------------------------------------------#

def draw_point_list(start,stop,pen,dc):
    """ Draw lines between the points listed in start and stop.
    
        start and stop are 2xN arrays of x,y point coordinates.  N lines
        are drawn, one between each of the start,stop pairs using the
        specified pen.  The lines are drawn on the given device context, dc.    
    """    
    dc.SetPen(pen)
    for i in range(len(start)):
        pt1 = start[i]
        pt2 = stop[i]
        dc.DrawLine(pt1[0],pt1[1],pt2[0],pt2[1])
    dc.SetPen(wx.wxNullPen)

#-----------------------------------------------------------------------#
#-------------------------- text objects -------------------------------#
#-----------------------------------------------------------------------#

class text_object(box_object,property_object):
    """ Text objects can be placed anywhere in the plot window.  They have 
        "font", "color", and "rotate" attributes as well as "visible" 
        attribute that specifies whether the text is displayed.  The font
        attribute is a wxWindows font.  "color" can either be a named
        color (string such as 'black') or a 3-element sequence specifiy
        rgb values (0 to 255).  Rotatation is specified in degrees.  
        Currently only 0 and 90 degree rotation are supported.
        
        Issues:
          -- The font attribute seems to be ignored when drawing
             rotated text. (wxPython)
          -- Need to think about origin when drawing rotated text.
             The current approach assumes the box_object does not
             rotate as the text rotates, so the insertion point changes
             from the topleft to bottomleft corner.  This inconsistency
             may be problematic in the future, but it was convienient
             for laying out axis titles.
          -- There is some ugliness with calculating the size of 
             the object because text needs to know the dc it is drawing
             in to know its size.  As a result, you must call set_dc(dc)
             before calling size.  To keep from having a dc reference
             hanging around, I added a clear_dc() method.  Necessary??   
    """    
    _attributes = {'color': ['black',colors,"Color of line"],
                   'font':  [None,[],"Font of text object"],
                   'rotate': [0,[], "Angular rotation of text in degrees"],
                   'visible': ['yes',[], "'yes' or 'no'. Is text visible"],
                   }
    #think about rotation and topleft
    def __init__(self,text=None,topleft=None,**attr):

        property_object.__init__(self,attr)
        box_object.__init__(self, (0,0), (0,0))
        if text is None: self.text = ''
        else: self.text = text
        if not topleft: self.topleft = array((0,0),TYPE)
        else:           self.topleft = array(topleft,TYPE)
        self.translate((0,0))
        # make sure each object has its own font object.
        if not self.font: self.font = default_font()
        if not self.color: self.color = 'black'

    # ugly hack to get size calculations to work
    def set_dc(self,dc):
        """ Set the device context of the text_object for size calculations.
        """
        self.dc = dc
    def clear_dc(self):
        """ Get rid of the device context.
        """
        #try/except cludge added right before SPIE conference
        #find error and fix.
        try:
            del self.dc
        except:
            pass
            
    def size(self):
        """ Calculate the width and length of the text in pixels.
        
            If "visible" == 'no', size() always returns (0,0) regardless
            of the text size.  If the rotation angle is 90 degrees,
            the width of the text is actually its height on the screen
            and visa-versa.
            
            Issues:
               -- haven't really thought about how to handle non-orthogonal
                  text rotations.
        """
        if self.visible in ['yes','on',1] and self.text:
            if not hasattr(self,'dc'):
                raise ValueError, "no device context to calculate text " \
                                  "size. Call set_dc() first."            
            preset = (self.dc.GetFont() == self.font)
            if not preset: self.dc.SetFont(self.font)
            sz= array(self.dc.GetTextExtent(self.text))
            
            # Commented the following line out since calling this
            # seems to turn of plotting of any text            
            #if not preset: self.dc.SetFont(wx.wxNullFont)
            
            # should do something here to calculate real width
            # this only works for 90 degree rotations
            if self.rotate: sz = sz[::-1]
        else:
            sz = array((0,0))
        return sz
                                                    
    def set_size(self):
        raise ValueError, "Can't set size of text objects"
        
    def draw(self,dc):
        """ Draw the text on the screen.
        
            If the text is rotated by 90 degrees, the "insertion point"
            is the bottom-left corner.  Otherwise the stanard top-left
            corner is used.
        """
        if self.visible in ['yes','on',1] and self.text:
            self.set_dc(dc)
            color = get_color(self.color)
            dc.SetTextForeground(color)
            dc.SetFont(self.font)
            # hmmm.  rotated text isn't being drawn with the correct font
            # also need to think about origin for rotated text
            if self.rotate == 90:
                ##dc.DrawRotatedText(self.text,self.left(),self.bottom(),
                ##                   self.rotate)    # mod by GAP 26092003
                dc.DrawRotatedText(self.text,int(self.left()),int(self.bottom()),
                                   int(self.rotate))
            else:
                dc.DrawText(self.text,self.left(),self.top()) 

            dc.SetPen(wx.wxNullPen)
            dc.SetFont(wx.wxNullFont)
            # hmmm. really should clear the dc...
            self.clear_dc()
            

class text_window(wx.wxWindow,text_object):
    """
    Window is currently size=0,0 window for event handling.  Text is
    NOT drawn in the window.  Maybe will do this later, but not sure
    if it is worth handling drawing th background (does GTK support
    transparent windows?).
    """
    def __init__(self,parent,text=None,topleft=None,**attr):
        self.plot_canvas = parent
        wx.wxWindow.__init__(self,parent,-1)
        #not quite right with attr - 1.6 makes this cleaner
        text_object.__init__(self,text,topleft)
        self.SetSize((0,0))

    def format_popup(self,pos):
        menu = wx.wxMenu()
        menu.Append(500, 'Change Text', 'Change Text')
        wx.EVT_MENU(self, 500, self.OnText)
        menu.Append(600, 'Change Font', 'Change Text Font')
        wx.EVT_MENU(self, 600, self.OnFont)
        #print 'fp:',self.text
        menu.UpdateUI()
        self.PopupMenuXY(menu,pos[0],pos[1])
        
    def OnFont(self,event):
        data = wx.wxFontData()
        data.SetColour(get_color(self.color))
        data.SetInitialFont(self.font)
        dlg = wx.wxFontDialog(self, data)
        if dlg.ShowModal() == wx.wxID_OK:
            data = dlg.GetFontData()
            self.font = data.GetChosenFont()
            color = data.GetColour()
            self.color = color.Red(),color.Green(),color.Blue()
            self.plot_canvas.update()
        dlg.Destroy()

    def OnText(self,event):
        #print 'on text:',self.text
        dlg = wx.wxTextEntryDialog(self.GetParent(), 'New Text',
                            defaultValue=self.text)
        if dlg.ShowModal() == wx.wxID_OK:
            self.text = dlg.GetValue()
            self.plot_canvas.update()     
        dlg.Destroy()

        
#-----------------------------------------------------------------------#
#-------------------------- legend_object ------------------------------#
#-----------------------------------------------------------------------#

class legend_object(property_object,box_object):
    """ Draws a legend to the lines plotted on the graph.
    
        Unimplemented.
    """
    _attributes = {
    'position': ['auto',[],"Specify location for legend - auto only for now"],
    'font':       [None,[],"Font of tick labels"],
    'border':     ['yes',['yes','no'],"Place border around legend"],
    'fill_color': ['white',colors,"Background color in legend"],
    'line_length':[30,[],"Length of lines displayed in legend"],
    'visible':    ['yes',['yes','on','no','off'],"Show the legend"],
    }
    def __init__(self, **attr):
        property_object.__init__(self,attr)
        box_object.__init__(self,(0,0),(0,0))
    def layout(self,graph_lines,graph_area,dc):
        
        self.graph_lines = graph_lines # keep a pointer so we can clone
                                       # properties when drawing
        # create duplicates of the lines            
        
        #figure out where to put it
        #create legend
    def draw(self,dc):
        if self.visible in ['yes','on']:
            # beginning of junk that should be moved
            # in to a line object
            self.labels = []
            self.lines = []
            self.markers = []
            
            for line in self.graph_lines:
                #self.labels.append(text_object(line.name))
                self.labels.append(text_object('bug'))
                self.labels[-1].set_dc(dc)
                #create legend lines and markers and steal properties
                #from actual line.  We'll update the points later
    
            # figure out size
            w,h = 0,0
            margin = 0
            line_pts = []
            line_right = point_object((self.line_length+5,0))
            previous = box_object((0,0),(0,0)) #a 0x0 box at the origin
            for i in range(len(self.labels)):
                label = self.labels[i]
                label.below(previous,margin)
                label.right_of(line_right)
                new_line = poly_line([(0,label.center_y()),
                                      (self.line_length,label.center_y())])
                new_line.clone_properties(line.line)
                self.lines.append(new_line)
                new_marker = poly_marker([(self.line_length/2.,
                                            label.center_y())])
                new_marker.clone_properties(line.markers)
                self.markers.append(new_marker)
                previous = label
    
            for items in self.labels: items.clear_dc()
            # end of junk that should be moved
            for item in self.lines:
                item.draw(dc)
            for item in self.markers:
                item.draw(dc)                
            for item in self.labels:
                item.draw(dc)

#-----------------------------------------------------------------------#
#---------------------------- axis_object ------------------------------#
#-----------------------------------------------------------------------#

class axis_object(property_object):
    """ Class for drawing an axis on a graph.
    
        
    """
    _attributes = {
    'labels_visible':  ['yes',['yes','no'], "Turns on/off labels for axis"],
    'label_color': ['black',colors, "Color of tick labels"],
    'label_font':  [None,[],"Font of tick labels"],
    'label_location':  ['minus',['plus','minus'],"Side of axis for labels"],
    'label_offset':  [4,[],"Offset in pixels from tick marks to tick labels"],
    'ticks_visible': ['yes',['yes','no'], "Turns on/off tick marks for axis"],
    'tick_in':    [5,tick_length, "Tick length in pixels into the graph"],
    'tick_out':   [0,tick_length, "Tick length in pixels out ofthe graph"],
    'tick_weight':[1,line_widths, "Width of tick lines"],
    'tick_color': ['black',colors,"Color of tick labels"],
    'tick_style': ['solid',line_styles,"Tick mark and axis line style"],
    'tick_interval': ['auto',[],"Tick interval: 'auto', 'linear','log', " \
                                "or value"],
    'bounds': [['auto','auto'],[],"Axis boundary: 'auto', 'fit', or value"],
    'grid_visible':  ['yes',['yes','no'], "Turns on and off grid lines" \
                                          " for axis"],
    'grid_color': ['light grey',colors,"Color for grid lines"],
    'grid_weight':[1,line_widths, "Width of grid lines"],
    'grid_style': ['dot',line_styles,"Line style for grid lines"],
                   }
    # axis crossing
    # others...
                    
                    
    def __init__(self, rotate = 0, graph_location='above',**attr):
        # attr to allow or disallow labels        
        property_object.__init__(self,attr)
        self.rotate = rotate
        if graph_location == 'above' or graph_location == 'left':
            self.tick_sign = -1
        elif graph_location == 'below' or graph_location == 'right':
            self.tick_sign = 1                        
        else: 
            raise ValueError, "graph_location can be 'left','right','above" \
                              ", or 'below'. You tried '%s'" % graph_location    
        if not self.label_font:
            self.label_font = default_font()
            self.label_font.SetPointSize(10)
        self.bounds = ['auto', 'auto'] # each object needs its own bounds list
        self.omit_first_label = 0
        self.omit_last_label = 0
    def calculate_ticks(self,data_bounds):
        """ data bounds is (lower bound, upper bound)
           axis_settings is a 3-tuple (lower bound, upper bound, interval)
           Each of its settings can be 'auto', or a numerical value.
           In addition, the bounds can be 'fit' and interval can be 
           'linear' or 'log' (well, not yet).
           'auto' -- bound or interval is calculated for you based on the 
                     data bounds.
           'fit'  -- bound is set to be slightly beyond than data_bounds
            value  -- bound is exactly to specified value.
           Typical settings are the following:
            ['auto','auto','auto'
        """    
        bounds_settings = [self.bounds[0], self.bounds[1], self.tick_interval]
        self.ticks = auto_ticks(data_bounds,bounds_settings)
        # This code tries to omit printing of first and last labels when they 
        # are strange values due to 'fit' settings.
        self.omit_first_label = 0
        self.omit_last_label = 0
        if len(self.ticks) > 3:
            delta = self.ticks[2] - self.ticks[1]
            first_delta = self.ticks[1] - self.ticks[0]
            last_delta = self.ticks[-1] - self.ticks[-2]
            first_diff = abs(first_delta - delta)
            last_diff = abs(last_delta - delta)
            if first_diff > .01*delta:
                self.omit_first_label = 1
            if last_diff > .01*delta:
                self.omit_last_label = 1
            
        self.create_labels()
    
    def create_labels(self):
        self.labels = []
        tick_text = format_tick_labels(self.ticks)
        for text in tick_text:
            label = text_object(text,(0,0),font=self.label_font,
                                           color=self.label_color)
            self.labels.append(label)
        if len(self.labels):    
            if self.omit_first_label:
                self.labels[0] = text_object(' ',(0,0),font=self.label_font,
                                               color=self.label_color)
            if self.omit_last_label:
                self.labels[-1] = text_object(' ',(0,0),font=self.label_font,
                                               color=self.label_color)
        
    
    def max_label_size(self,dc,idx):        
        dim = [0]
        for i in self.labels:
            i.set_dc(dc)
            dim.append(i.size()[idx])
            i.clear_dc()
        return max(dim) + self.tick_out + self.label_offset

    def max_label_width(self,dc):        
        return self.max_label_size(dc,0)
    def max_label_height(self,dc):        
        return self.max_label_size(dc,1)
    def range(self):
        return float(self.ticks[-1]-self.ticks[0])
        
    def layout(self,graph_area,dc):
        if self.rotate == 0:           
            length = graph_area.width()
            grid_length = graph_area.height()
        elif self.rotate == 90 or self.rotate == -90: 
            length = graph_area.height()
            grid_length = graph_area.width()
        else: 
            raise ValueError,'rotate must be 0 or 90. It is %d' % self.rotate            

        # translate to screen units for the ticks.
        scale = length / float(self.ticks[-1] - self.ticks[0])
        tick_locations = floor((self.ticks - self.ticks[0]) * scale)
        
        # make array of points with tick screen units as x value, 0 for y value
        tick_points = []
        for i in tick_locations:
            tick_points.append((i,0))            
        tick_points = array(tick_points)

        # create array of end points for the tick lines
        tick_start =  tick_points + array((0,self.tick_in*self.tick_sign))
        tick_stop =  tick_points - array((0,self.tick_out*self.tick_sign))            
        
        # set up grid lines
        grid_start = tick_points 
        grid_stop =  tick_points + array((0,grid_length*self.tick_sign))
        
        # rotate all of this stuff to the correct axis orientation.        
        angle = -self.rotate*pi/180 # neg. since up is y decreasing in canvas.
        zero_point = array((0,0))
        self.tick_points = rotate(tick_points,zero_point,angle)
        self.tick_start = rotate(tick_start,zero_point,angle)
        self.tick_stop = rotate(tick_stop,zero_point,angle)
        self.grid_start = rotate(grid_start,zero_point,angle)
        self.grid_stop = rotate(grid_stop,zero_point,angle)
        #offset the labels from the tick_points
        if self.label_location == 'plus':
            if tan(self.rotate*pi/180) < 1: label_angle = angle - pi/2
            else: label_angle = angle + pi/2
        elif self.label_location == 'minus':
            if tan(self.rotate*pi/180) < 1: label_angle = angle + pi/2
            else: label_angle = angle - pi/2
        for i in range(len(self.labels)):
            self.labels[i].set_dc(dc)
            pt = point_object(self.tick_points[i])        
            self.labels[i].radial_offset_from(pt,label_angle,
                                              margin=self.label_offset)
            self.labels[i].clear_dc()
        self.labels[i].set_dc(dc)            

        # for fast line drawing.
        # not currently used
        self.single_grid_line = []
        self.single_tick_line = []
        ts = map(tuple,self.tick_start)
        tc = map(tuple,self.tick_points)
        te = map(tuple,self.tick_stop)
        N = len(self.tick_start)
        for i in range(N):
            self.single_tick_line.append(tc[i])
            self.single_tick_line.append(te[i])
            self.single_tick_line.append(ts[i])
            self.single_tick_line.append(tc[i])

        
    def move(self,offset):
        self.tick_start = self.tick_start + offset
        self.tick_stop = self.tick_stop + offset
        self.tick_points = self.tick_points + offset
        self.grid_start = self.grid_start + offset
        self.grid_stop = self.grid_stop + offset
        # need to rebuild fast_line drawing stuff
        for i in self.labels:
            i.translate(offset)
    
    def contains(self,pt,dc):
        mins,maxs = [],[]
        for i in self.labels: i.set_dc(dc)
        lo,hi = bounding_points(self.labels)
        for i in self.labels: i.clear_dc()
        mins.append(lo)
        mins.append(minimum.reduce(self.tick_start))
        mins.append(minimum.reduce(self.tick_stop))
        mins.append(minimum.reduce(self.tick_points))
        maxs.append(hi)
        maxs.append(maximum.reduce(self.tick_start))
        maxs.append(maximum.reduce(self.tick_stop))
        maxs.append(maximum.reduce(self.tick_points))
        lo = minimum.reduce(mins)
        hi = maximum.reduce(maxs)
        sz = hi-lo
        return box_object(lo,sz).contains(pt)
          
    def draw_ticks(self,dc):
        if self.ticks_visible in ['yes','on']:
            style = line_style_map[self.tick_style]
            color = get_color(self.tick_color)
            pen = wx.wxPen(color, self.tick_weight, style)        
            draw_point_list(self.tick_start,self.tick_stop,pen,dc)
            #dc.DrawLines(self.single_tick_line)
        # draw axis line here also
        pt1 = self.tick_points[0]    
        pt2 = self.tick_points[-1]
        draw_point_list([pt1],[pt2],pen,dc)
        
    def draw_grid_lines(self,dc):
        if self.grid_visible in ['yes','on']:
            style = line_style_map[self.grid_style]
            color = get_color(self.grid_color)
            pen = wx.wxPen(color,self.grid_weight, style)        
            draw_point_list(self.grid_start,self.grid_stop,pen,dc)

    def draw_labels(self,dc):
        if self.labels_visible in ['yes','on']:
            for label in self.labels:
                label.draw(dc)
                    
    def draw_lines(self,dc):
        self.draw_grid_lines(dc)
        self.draw_ticks(dc)

    def draw(self,dc):
        self.draw_labels(dc) # really only need to do this sometimes...
        self.draw_lines(dc)

class axis_window(wx.wxWindow,axis_object):
    """
    Window is currently size=0,0 window for event handling.  Text is
    NOT drawn in the window.  Maybe will do this later, but not sure
    if it is worth handling drawing th background (does GTK support
    transparent windows?).
    """
    def __init__(self,parent, rotate = 0, graph_location='above',**attr):
        self.plot_canvas = parent
        wx.wxWindow.__init__(self,parent,-1)
        #not quite right with attr - 1.6 makes this cleaner
        axis_object.__init__(self,rotate,graph_location)
        self.SetSize((0,0))

    def format_popup(self,pos):
        menu = wx.wxMenu()
        menu.Append(600, 'Change Font', 'Change Text Font')
        wx.EVT_MENU(self, 600, self.OnFont)
        menu.UpdateUI()
        self.PopupMenuXY(menu,pos[0],pos[1])
        
    def OnFont(self,event):
        data = wx.wxFontData()
        data.SetColour(get_color(self.label_color))
        data.SetInitialFont(self.label_font)
        dlg = wx.wxFontDialog(self, data)
        if dlg.ShowModal() == wx.wxID_OK:
            data = dlg.GetFontData()
            self.label_font = data.GetChosenFont()
            color = data.GetColour()
            self.label_color = color.Red(),color.Green(),color.Blue()
            self.plot_canvas.update()
        dlg.Destroy()

#-----------------------------------------------------------------------#
#-------------------------- border_object ------------------------------#
#-----------------------------------------------------------------------#

class border_object(property_object):
    _attributes = {
    'ticks_visible': ['yes',['yes','no'], "Turns on/off tick marks for axis"],
    'tick_in':    [5,tick_length, "Tick length in pixels into the graph"],
    'tick_out':   [0,tick_length, "Tick length in pixels out ofthe graph"],
    'tick_weight':[1,line_widths, "Width of tick lines"],
    'tick_color': ['black',colors,"Color of tick labels"],
    'tick_style': ['solid',line_styles,"Tick mark line style"],
    'visible':  ['yes',['yes','no'], "Turns on and off border marks for axis"],
    'color': ['black',colors,"Color for border lines"],
    'weight':[1,line_widths, "Width of border lines"],
    'style': ['solid',line_styles,"Line style for border lines"],
    }
    def __init__(self,**attr):
        # attr to allow or disallow labels        
        property_object.__init__(self,attr)
    
    def add_ticks(self,tick_points, origin, tick_in, tick_out):
        offset = array((0,self.tick_out))
        self.tick_start.append( tick_points + offset + origin )
        offset = array((0,self.tick_in))
        self.tick_stop.append( tick_points + offset + origin )

    def layout(self,graph_area,x_axis,y_axis):
        # top
        x_points = x_axis.tick_points - x_axis.tick_points[0]
        y_points = y_axis.tick_points - y_axis.tick_points[0]
        
        #clear tick lines
        self.tick_start = []
        self.tick_stop = []
        self.tick_center = []
        
        #top
        origin = array((graph_area.left(),graph_area.top()))
        offset = array((0,-self.tick_out))
        self.tick_start.append( x_points + offset + origin )
        offset = array((0,self.tick_in))
        self.tick_stop.append( x_points + offset + origin )
        self.tick_center.append(x_points + origin)
        #right
        origin = array((graph_area.right(),graph_area.bottom()))
        offset = array((self.tick_out,0))
        self.tick_start.append( y_points + offset + origin )
        offset = array((-self.tick_in,0))
        self.tick_stop.append( y_points + offset + origin )
        self.tick_center.append(y_points + origin)
        #bottom
        origin = array((graph_area.left(),graph_area.bottom()))
        offset = array((0,self.tick_out))
        self.tick_start.append( x_points + offset + origin )
        offset = array((0,-self.tick_in))
        self.tick_stop.append( x_points + offset + origin )
        self.tick_center.append(x_points + origin)
        #left
        origin = array((graph_area.left(),graph_area.bottom()))
        offset = array((-self.tick_out,0))
        self.tick_start.append( y_points + offset + origin )
        offset = array((self.tick_in,0))
        self.tick_stop.append( y_points + offset + origin )
        self.tick_center.append(y_points + origin)
        
        self.tick_start = concatenate(self.tick_start)
        self.tick_stop = concatenate(self.tick_stop)
        self.tick_center = concatenate(self.tick_center)
        
        self.border_start = []
        self.border_stop = []
        gr = graph_area
        # could do a rect instead, but it always seems
        # off by 1 or so. Do this for now.
        self.border_start.append((gr.left(),gr.top()))
        self.border_stop.append((gr.left(),gr.bottom()))
        self.border_start.append((gr.left(),gr.top()))
        self.border_stop.append((gr.right(),gr.top()))
        self.border_start.append((gr.left(),gr.bottom()))
        self.border_stop.append((gr.right(),gr.bottom()))
        self.border_start.append((gr.right(),gr.top()))
        self.border_stop.append((gr.right(),gr.bottom()))
        self.border_start = array(self.border_start)
        self.border_stop = array(self.border_stop)
                
        # this stuff is only needed for the fast_draw method
        # needs slight amount of work
        self.single_line = []
        ts = map(tuple,self.tick_start)
        tc = map(tuple,self.tick_center)
        te = map(tuple,self.tick_stop)
        N = len(self.tick_start)
        for i in range(N):
            self.single_line.append(tc[i])
            self.single_line.append(te[i])
            self.single_line.append(ts[i])
            self.single_line.append(tc[i])
            
    def draw(self,dc):
        vis = self.visible in ['yes','on']
        tick_vis = self.ticks_visible in ['yes','on']
        
        #if vis and tick_vis:
        if 0:
            # Really should check styles etc. here to make
            # sure the are the same.
            self.draw_fast(dc)
        else:    
            #draw border
            if vis:
                style = line_style_map[self.style]
                color = get_color(self.color)
                pen = wx.wxPen(color, self.weight, style)        
                draw_point_list(self.border_start,self.border_stop,pen,dc)
            #draw ticks
            if tick_vis:
                style = line_style_map[self.tick_style]
                color = get_color(self.tick_color)
                pen = wx.wxPen(color, self.tick_weight, style)        
                draw_point_list(self.tick_start,self.tick_stop,pen,dc)
        
    def draw_fast(self,dc):
        """ This approach uses wxPythons DrawLines to draw
            the entire border in one call instead of drawing
            each tick individually.  It is 5-10 times faster,
            but draws the border and ticks all in the same style.
            (not a big draw back...)
        """
        style = line_style_map[self.style]
        color = get_color(self.color)
        pen = wx.wxPen(color, self.weight, style)        
        dc.SetPen(pen)
        dc.DrawLines(self.single_line)
        dc.SetPen(wx.wxNullPen)

class poly_points(property_object):
    def __init__(self, points, attr=None):
        property_object.__init__(self,attr)
        self.points = array(points)
        self.scaled = array(self.points,copy=1)
        #self.scaled = map(tuple,self.points)

    def bounding_box(self):
        return minimum.reduce(self.points), \
               maximum.reduce(self.points)

    def scale_and_shift(self, scale=1, shift=0):
        self.scaled = scale*self.points+shift
        #self.scaled = map(tuple,self.scaled)
       
class poly_line(poly_points):
    _attributes = {'color': ['black',colors,"Color of line"],
                   'weight': [1,line_widths, "Weight of Line"],
                   'visible':    ['yes',['yes','no'], "Turn on/off line"],
                   'style': ['solid',line_styles,"solid, dash, dot dash, or dot"]
                  }

    def __init__(self, points, **attr):
        poly_points.__init__(self, points, attr)
    def draw(self, dc):
        if self.visible in ['on','yes']:
            color = get_color(self.color)
            style = line_style_map[self.style]
            dc.SetPen(wx.wxPen(color, self.weight,style))
            try:
                dc.DrawLines(self.scaled)           
            except:
                dc.DrawLines(map(tuple,self.scaled))
            dc.SetPen(wx.wxNullPen)
                    
marker_styles = ['circle','square','dot','triangle','down_triangle',\
                 'cross','plus']                    

class poly_marker(poly_points):
    """ adding a layout method that did most of the computation
        and marker function selection upfront might speed up drawing
    """
    _attributes = {
       'outline_color': ['black',colors,"Color of marker outline"],
       'outline_weight': [1,line_widths, "Weight of Line"],
       'fill_color': ['black',colors,"Fill color of marker"],
       'fill_style': ['solid',fill_styles, "pattern used when filling marker"],
       'size':       [1,[],"Size of marker"],
       'symbol':     ['circle',marker_styles,"Shape used for marker"],
       'visible':    ['yes',['yes','no'], "Turn on/off markers"],       
       }
    def __init__(self,points,**attr):
        poly_points.__init__(self,points,attr)

    def draw(self,dc):
        if self.visible in ['on','yes']:
            color = get_color(self.outline_color)
            weight = self.outline_weight
            size = self.size
            fill_color = get_color(self.fill_color)
            fill_style = fill_style_map[self.fill_style]
            symbol = self.symbol
            dc.SetPen(wx.wxPen(color,weight))
            dc.SetBrush(wx.wxBrush(fill_color,fill_style))
            self._drawmarkers(dc, self.scaled, symbol, size)
            dc.SetPen(wx.wxNullPen)
            dc.SetBrush(wx.wxNullBrush)
    def _drawmarkers(self, dc, coords, marker,size=1):
        f = eval('self._' +marker)
        for xc, yc in coords:
            f(dc, xc, yc, size)

    def _circle(self, dc, xc, yc, size=1):
        ##dc.DrawEllipse(xc-3*size,yc-3*size,6*size,6*size)   #mod by GAP 26092003
        dc.DrawEllipse(int(xc-3*size),int(yc-3*size),6*size,6*size)

    def _dot(self, dc, xc, yc, size=1):
        dc.DrawPoint(xc,yc)

    def _square(self, dc, xc, yc, size=1):
        dc.DrawRectangle(xc-3*size,yc-3*size,6*size,6*size)

    def _triangle(self, dc, xc, yc, size=1):
        dc.DrawPolygon([(-0.5*size*6,0.2886751*size*6),
                        (0.5*size*6,0.2886751*size*6),
                        (0.0,-0.577350*size*6)],xc,yc)

    def _down_triangle(self, dc, xc, yc, size=1):
        dc.DrawPolygon([(-0.5*size*6,-0.2886751*size*6),
                        (0.5*size*6,-0.2886751*size*6),
                        (0.0,0.577350*size*6)],xc,yc)

    def _cross(self, dc, xc, yc, size=1):
        dc.DrawLine(xc-3*size,yc-3*size,xc+3*size,yc+3*size)
        dc.DrawLine(xc-3*size,yc+3*size,xc+3*size,yc-3*size)

    def _plus(self, dc, xc, yc, size=1):
        dc.DrawLine(xc-3*size,yc,xc+3*size,yc)
        dc.DrawLine(xc,yc-3*size,xc,yc+3*size)

class line_object(poly_points):
    """ Combines poly_line and poly_marker into a single
        class
    """
    _attributes = {
       'color': ['auto',['auto','custom'],"'auto' or 'custom'."],
       'marker_type':['auto',['auto','custom'],"'auto' or 'custom'."],
       'line_type': ['auto',['auto','custom'],"'auto' or 'custom'."],
       'name': ['','text string that identifies line'],
       }

    def __init__(self,points,**attr):
        poly_points.__init__(self,points,attr)
        self.markers = poly_marker(points)
        self.line = poly_line(points)
    def clip_box(self,box):
        self.clip = box.left(),box.top(),box.width(),box.height()
    def draw(self,dc):                   
        if hasattr(self,'clip'):
            c = self.clip
            ##dc.SetClippingRegion(c[0]-1,c[1]-1,c[2]+2,c[3]+2) # mod by GAP 26092003
            dc.SetClippingRegion(int(c[0]-1),int(c[1]-1),int(c[2]+2),int(c[3]+2))
        self.line.draw(dc)
        if hasattr(self,'clip'): dc.DestroyClippingRegion()
        if hasattr(self,'clip'):
            c = self.clip
            ##dc.SetClippingRegion(c[0]-5,c[1]-5,c[2]+10,c[3]+10)  # mod by GAP 26092003
            dc.SetClippingRegion(int(c[0]-5),int(c[1]-5),int(c[2]+10),int(c[3]+10))
        self.markers.draw(dc)        
        if hasattr(self,'clip'): dc.DestroyClippingRegion()
        
    def scale_and_shift(self, scale=1, shift=0):
        self.markers.scale_and_shift(scale, shift)
        self.line.scale_and_shift(scale, shift)
    def set_color(self,color):
        self.line.color = color
        self.markers.outline_color = color
        self.markers.fill_color = color


        
class image_object(property_object):
    _attributes = {
      'colormap': ['grey',[],"name of colormap or Nx3 color array"],
      'scale':    ['yes',['yes','no'],"Turn scaling on/off"],
      }
    def __init__(self, matrix,x_bounds=None,y_bounds=None,**attr):
        property_object.__init__(self,attr)        
        if not x_bounds: 
            self.x_bounds = array((0,matrix.shape[1]))
        else:
            # works for both 2 element or N element x
            self.x_bounds = array((x_bounds[0],x_bounds[-1]))
        if not y_bounds: 
            self.y_bounds = array((0,matrix.shape[0]))
        else:
            self.y_bounds = array((y_bounds[0],y_bounds[-1]))
        
        self.matrix = matrix
        self.the_image = self.form_image()

        axis_lengths = array((self.x_bounds[1]-self.x_bounds[0],
                              self.y_bounds[1]-self.y_bounds[0]))
        self.image_pixels_per_axis_unit =array((matrix.shape[1], matrix.shape[0]),Float)/axis_lengths
        self.image_origin = array((self.x_bounds[0],self.y_bounds[0]))
        
    def scale_magnitude(self,image,colormap):
        themin = float(minimum.reduce(ravel(image)))
        themax = float(maximum.reduce(ravel(image)))
        scaled = (image - themin) / (themax-themin) * len(colormap) *.99999
        scaled = scaled.astype('b')
        return scaled
        
    def form_image(self):
        # look up colormap if it si identified by a string
        if type(self.colormap) == type(''):
            try:
                colormap = colormap_map[self.colormap]
            except KeyError:
                raise KeyError, 'Invalid colormap name. Choose from %s' \
                                % `colormap_map.keys()`
        else:
            colormap = self.colormap
        # scale image if we're supposed to.    
        if self.scale in ['yes','on']:
            scaled_mag = self.scale_magnitude(self.matrix,colormap)
        else:    
            scaled_mag = self.matrix.astype('b')
        scaled_mag = clip(scaled_mag,0,len(colormap)-1)
        
        if float(maximum.reduce(ravel(colormap))) == 1.:
            cmap = colormap * 255
        else:
            cmap = colormap    
        
        pixels = take( cmap, scaled_mag)
        del scaled_mag
        # need to transpose pixels in memory...
        bitmap = pixels.astype(UnsignedInt8).tostring()
        image = wx.wxEmptyImage(self.matrix.shape[1],self.matrix.shape[0])
        image.SetData(bitmap)
        return image
        
    def bounding_box(self):
        bb = array((self.x_bounds[0],self.y_bounds[0])), \
             array((self.x_bounds[1],self.y_bounds[1]))
        return bb

    def scale_and_shift(self, scale=1, shift=0,upperleft=None,size=None):
        if scale is 1: scale = array((1,1))
        if shift is 0: shift = array((0,0))
        graph_pixels_per_axis_unit = scale
        self.scale = graph_pixels_per_axis_unit/self.image_pixels_per_axis_unit
        self.origin = shift
        self.graph_upperleft = upperleft
        self.graph_size = size

    def draw(self,dc):
        sz = array((self.the_image.GetWidth(),self.the_image.GetHeight()))
        sz = sz* self.scale
        sz = abs(sz.astype(Int))
        scaled_image = self.the_image.Scale(sz[0],sz[1])
        bitmap = scaled_image.ConvertToBitmap()

        dc.DrawBitmap(bitmap, self.origin[0]+1, 
                      self.origin[1]-scaled_image.GetHeight()+1, wx.false)

import UserList
huge = array((1e308,1e308))
tiny = array((-1e308,-1e308))
class graphic_list(UserList.UserList):
    """ probably should have a layout method
    """
    def __init__(self):
        UserList.UserList.__init__(self)
    
    def bounding_box(self):
        p1 =[]; p2 = []
        for o in self.data:
            p1o, p2o = o.bounding_box()
            p1.append(p1o);p2.append(p2o)
        if len(p1):
            return minimum.reduce(p1), maximum.reduce(p2)
        else: # so big or small that they always loose to other data sets
            return huge,tiny

    def axis_bounds(self):
        p1, p2 = self.bounding_box()        
        return array((p1[0],p2[0])), array((p1[1],p2[1]))
    
    def scale_and_shift(self, scale=1, shift=0):
        for o in self.data:
            o.scale_and_shift(scale, shift)

    def draw(self, canvas):
        for o in self.data:
            o.draw(canvas)

auto_styles = [['blue','circle','solid'], ['green','square','solid'],
               ['red','triangle','solid'],['cyan','down_triangle','solid'],
               ['orange','cross','solid'],['purple','plus','solid'] ]
              
# make a black and white auto?
      
class auto_line_list(graphic_list):
    """ Handles drawing of line_objects with auto styles
        only works with line_objects
    """
    def clip_box(self,box):
        for o in self.data:
            o.clip_box(box)
    def draw(self, canvas):
        idx = 0
        for o in self.data:
            #print 'line vis before:',o.line_type, o.line.visible
            #print 'marker vis before:',o.marker_type, o.markers.visible
            if 'auto' in [o.color,o.line_type,o.marker_type]:
                if idx >= len(auto_styles): 
                    idx = 0
                color,marker,line_type = auto_styles[idx]
                #print 'clm: ',color,marker,line_type
                if 'auto' == o.color: o.set_color(color)
                if 'auto' == o.marker_type and marker:
                    o.markers.symbol = marker
                    o.markers.visible = 'yes'
                if 'auto' == o.line_type and line_type:
                    o.line.style = line_type
                    o.line.visible = 'yes'
                idx = idx + 1
            #print 'line vis after:',o.line_type, o.line.visible
            #print 'marker vis after:',o.marker_type, o.markers.visible
            o.draw(canvas)          
